"""Miscallaneous utility functions for the psyplot package"""
import re
import six
from difflib import get_close_matches
from itertools import chain
from psyplot.compat.pycompat import OrderedDict, filterfalse
from psyplot.docstring import dedent, docstrings


class DefaultOrderedDict(OrderedDict):
    """An ordered :class:`collections.defaultdict`

    Taken from http://stackoverflow.com/a/6190500/562769"""
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and
           not callable(default_factory)):
            raise TypeError('first argument must be callable')
        OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()

    def copy(self):
        """Return a shallow copy of the dictionary"""
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy
        return type(self)(self.default_factory,
                          copy.deepcopy(self.items()))

    def __repr__(self):
        return 'DefaultOrderedDict(%s, %s)' % (self.default_factory,
                                               OrderedDict.__repr__(self))


class _TempBool(object):
    """Wrapper around a boolean defining an __enter__ and __exit__ method

    Notes
    -----
    If you want to use this class as an instance property, rather use the
    :func:`_temp_bool_prop` because this class as a descriptor is ment to be a
    class descriptor"""

    #: default boolean value for the :attr:`value` attribute
    default = False

    #: boolean value indicating whether there shall be a validation or not
    value = False

    def __init__(self, default=False):
        """
        Parameters
        ----------
        default: bool
            value of the object"""
        self.default = default
        self.value = default
        self._entered = []

    def __enter__(self):
        self.value = not self.default
        self._entered.append(1)

    def __exit__(self, type, value, tb):
        self._entered.pop(-1)
        if not self._entered:
            self.value = self.default

    if six.PY2:
        def __nonzero__(self):
            return self.value
    else:
        def __bool__(self):
            return self.value

    def __repr__(self):
        return repr(bool(self))

    def __str__(self):
        return str(bool(self))

    def __call__(self, value=None):
        """
        Parameters
        ----------
        value: bool or None
            If None, the current value will be negated. Otherwise the current
            value of this instance is set to the given `value`"""
        if value is None:
            self.value = not self.value
        else:
            self.value = value

    def __get__(self, instance, owner):
        return self

    def __set__(self, instance, value):
        self.value = value


def _temp_bool_prop(propname, doc="", default=False):
    """Creates a property that uses the :class:`_TempBool` class

    Parameters
    ----------
    propname: str
        The attribute name to use. The _TempBool instance will be stored in the
        ``'_' + propname`` attribute of the corresponding instance
    doc: str
        The documentation of the property
    default: bool
        The default value of the _TempBool class"""
    def getx(self):
        if getattr(self, '_' + propname, None) is not None:
            return getattr(self, '_' + propname)
        else:
            setattr(self, '_' + propname, _TempBool(default))
        return getattr(self, '_' + propname)

    def setx(self, value):
        getattr(self, propname).value = bool(value)

    def delx(self):
        getattr(self, propname).value = default

    return property(getx, setx, delx, doc)


def unique_everseen(iterable, key=None):
    """List unique elements, preserving order. Remember all elements ever seen.

    Function taken from https://docs.python.org/2/library/itertools.html"""
    # unique_everseen('AAAABBBCCDAABBB') --> A B C D
    # unique_everseen('ABBCcAD', str.lower) --> A B C D
    seen = set()
    seen_add = seen.add
    if key is None:
        for element in filterfalse(seen.__contains__, iterable):
            seen_add(element)
            yield element
    else:
        for element in iterable:
            k = key(element)
            if k not in seen:
                seen_add(k)
                yield element


def is_remote_url(path):
    patt = re.compile(r'^https?\://')
    if not isinstance(path, six.string_types):
        return all(map(patt.search, (s or '' for s in path)))
    return bool(re.search(r'^https?\://', path))


@docstrings.get_sections(base='check_key', sections=['Parameters', 'Returns',
                                                 'Raises'])
@dedent
def check_key(key, possible_keys, raise_error=True,
              name='formatoption keyword',
              msg=("See show_fmtkeys function for possible formatopion "
                   "keywords"),
              *args, **kwargs):
    """
    Checks whether the key is in a list of possible keys

    This function checks whether the given `key` is in `possible_keys` and if
    not looks for similar sounding keys

    Parameters
    ----------
    key: str
        Key to check
    possible_keys: list of strings
        a list of possible keys to use
    raise_error: bool
        If not True, a list of similar keys is returned
    name: str
        The name of the key that shall be used in the error message
    msg: str
        The additional message that shall be used if no close match to
        key is found
    ``*args,**kwargs``
        They are passed to the :func:`difflib.get_close_matches` function
        (i.e. `n` to increase the number of returned similar keys and
        `cutoff` to change the sensibility)

    Returns
    -------
    str
        The `key` if it is a valid string, else an empty string
    list
        A list of similar formatoption strings (if found)
    str
        An error message which includes

    Raises
    ------
    KeyError
        If the key is not a valid formatoption and `raise_error` is True"""
    if key not in possible_keys:
        similarkeys = get_close_matches(key, possible_keys, *args, **kwargs)
        if similarkeys:
            msg = ('Unknown %s %s! Possible similiar '
                   'frasings are %s.') % (name, key, ', '.join(similarkeys))
        else:
            msg = ("Unknown %s %s! ") % (name, key) + msg
        if not raise_error:
            return '', similarkeys, msg
        raise KeyError(msg)
    else:
        return key, [key], ''


def sort_kwargs(kwargs, *param_lists):
    """Function to sort keyword arguments and sort them into dictionaries

    This function returns dictionaries that contain the keyword arguments
    from `kwargs` corresponding given iterables in ``*params``

    Parameters
    ----------
    kwargs: dict
        Original dictionary
    ``*param_lists``
        iterables of strings, each standing for a possible key in kwargs

    Returns
    -------
    list
        len(params) + 1 dictionaries. Each dictionary contains the items of
        `kwargs` corresponding to the specified list in ``*param_lists``. The
        last dictionary contains the remaining items"""
    return chain(
        ({key: kwargs.pop(key) for key in params.intersection(kwargs)}
         for params in map(set, param_lists)), [kwargs])


def hashable(val):
    """Test if `val` is hashable and if not, get it's string representation

    Parameters
    ----------
    val: object
        Any (possibly not hashable) python object

    Returns
    -------
    val or string
        The given `val` if it is hashable or it's string representation"""
    if val is None:
        return val
    try:
        hash(val)
    except TypeError:
        return repr(val)
    else:
        return val


@docstrings.get_sections(base='join_dicts')
def join_dicts(dicts, delimiter=None, keep_all=False):
    """Join multiple dictionaries into one

    Parameters
    ----------
    dicts: list of dict
        A list of dictionaries
    delimiter: str
        The string that shall be used as the delimiter in case that there
        are multiple values for one attribute in the arrays. If None, they
        will be returned as sets
    keep_all: bool
        If True, all formatoptions are kept. Otherwise only the intersection

    Returns
    -------
    dict
        The combined dictionary"""
    if not dicts:
        return {}
    if keep_all:
        all_keys = set(chain(*(d.keys() for d in dicts)))
    else:
        all_keys = set(dicts[0])
        for d in dicts[1:]:
            all_keys.intersection_update(d)
    ret = {}
    for key in all_keys:
        vals = {hashable(d.get(key, None)) for d in dicts} - {None}
        if len(vals) == 1:
            ret[key] = next(iter(vals))
        elif delimiter is None:
            ret[key] = vals
        else:
            ret[key] = delimiter.join(map(str, vals))
    return ret


def is_iterable(iterable):
    """Test if an object is iterable

    Parameters
    ----------
    iterable: object
        The object to test

    Returns
    -------
    bool
        True, if the object is an iterable object"""
    try:
        iter(iterable)
    except TypeError:
        return False
    else:
        return True



"""Core package for interactive visualization in the psyplot package

This package defines the :class:`Plotter` and :class:`Formatoption` classes,
the core of the visualization in the :mod:`psyplot` package. Each
:class:`Plotter` combines a set of formatoption keys where each formatoption
key is represented by a :class:`Formatoption` subclass."""

import six
import weakref
from abc import ABCMeta, abstractmethod
from textwrap import TextWrapper
import logging
from itertools import chain, groupby, tee, repeat, starmap
from collections import defaultdict
from threading import RLock
from datetime import datetime, timedelta
from numpy import datetime64, timedelta64, ndarray, inf
from xarray.core.formatting import format_timestamp, format_timedelta
from psyplot import rcParams
from psyplot.warning import warn, critical, PsyPlotRuntimeWarning
from psyplot.compat.pycompat import map, filter, zip, range
from psyplot.config.rcsetup import SubDict
from psyplot.docstring import docstrings, dedent
from psyplot.data import (
    InteractiveList, _no_auto_update_getter, CFDecoder)
from psyplot.utils import (DefaultOrderedDict, _TempBool, _temp_bool_prop,
                           unique_everseen, check_key)

#: the default function to use when printing formatoption infos (the default is
#: use print or in the gui, use the help explorer)
default_print_func = six.print_


#: :class:`dict`. Mapping from group to group names
groups = {
    'data': 'Data manipulation formatoptions',
    'axes': 'Axes formatoptions',
    'labels': 'Label formatoptions',
    'plotting': 'Plot formatoptions',
    'post_processing': 'Post processing formatoptions',
    'colors': 'Color coding formatoptions',
    'misc': 'Miscallaneous formatoptions',
    'ticks': 'Axis tick formatoptions',
    'vector': 'Vector plot formatoptions',
    'masking': 'Masking formatoptions',
    'regression': 'Fitting formatoptions',
    }


def _identity(*args):
    """identity function to make no validation

    Returns
    -------
    object
        just return the last argument in ``*args``"""
    return args[-1]


def format_time(x):
    """Formats date values

    This function formats :class:`datetime.datetime` and
    :class:`datetime.timedelta` objects (and the corresponding numpy objects)
    using the :func:`xarray.core.formatting.format_timestamp` and the
    :func:`xarray.core.formatting.format_timedelta` functions.

    Parameters
    ----------
    x: object
        The value to format. If not a time object, the value is returned

    Returns
    -------
    str or `x`
        Either the formatted time object or the initial `x`"""
    if isinstance(x, (datetime64, datetime)):
        return format_timestamp(x)
    elif isinstance(x, (timedelta64, timedelta)):
        return format_timedelta(x)
    elif isinstance(x, ndarray):
        return list(x) if x.ndim else x[()]
    return x


def is_data_dependent(fmto, data):
    """Check whether a formatoption is data dependent

    Parameters
    ----------
    fmto: Formatoption
        The :class:`Formatoption` instance to check
    data: xarray.DataArray
        The data array to use if the :attr:`~Formatoption.data_dependent`
        attribute is a callable

    Returns
    -------
    bool
        True, if the formatoption depends on the data"""
    if callable(fmto.data_dependent):
        return fmto.data_dependent(data)
    return fmto.data_dependent


def _child_property(childname):
    def get_x(self):
        return getattr(self.plotter, self._child_mapping[childname])

    return property(
        get_x, doc=childname + " Formatoption instance in the plotter")


class FormatoptionMeta(ABCMeta):
    """Meta class for formatoptions

    This class serves as a meta class for formatoptions and allows a more
    efficient docstring generation by using the
    :attr:`psyplot.docstring.docstrings` when creating a new formatoption
    class"""
    def __new__(cls, clsname, bases, dct):
        """Assign an automatic documentation to the formatoption"""
        doc = dct.get('__doc__')
        if doc is not None:
            dct['__doc__'] = docstrings.dedent(doc)
        new_cls = super(FormatoptionMeta, cls).__new__(cls, clsname, bases,
                                                       dct)
        for childname in chain(new_cls.children, new_cls.dependencies,
                               new_cls.connections, new_cls.parents):
            setattr(new_cls, childname, _child_property(childname))
        if new_cls.plot_fmt:
            new_cls.data_dependent = True
        return new_cls


# priority values

#: Priority value of formatoptions that are updated before the data is loaded.
START = 30
#: Priority value of formatoptions that are updated before the plot it made.
BEFOREPLOTTING = 20
#: Priority value of formatoptions that are updated at the end.
END = 10


@six.add_metaclass(FormatoptionMeta)
class Formatoption(object):
    """Abstract formatoption

    This class serves as an abstract version of an formatoption descriptor
    that can be used by :class:`~psyplot.plotter.Plotter` instances."""

    priority = END
    """:class:`int`. Priority value of the the formatoption determining when
    the formatoption is updated.

    - 10: at the end (for labels, etc.)
    - 20: before the plotting (e.g. for colormaps, etc.)
    - 30: before loading the data (e.g. for lonlatbox)"""

    #: :class:`str`. Formatoption key of this class in the
    #: :class:`~psyplot.plotter.Plotter` class
    key = None

    _plotter = None

    @property
    def plotter(self):
        """:class:`~psyplot.plotter.Plotter`. Plotter instance this
        formatoption belongs to"""
        if self._plotter is None:
            return
        return self._plotter()

    @plotter.setter
    def plotter(self, value):
        if value is not None:
            self._plotter = weakref.ref(value)
        else:
            self._plotter = value

    #: `list of str`. List of formatoptions that have to be updated before this
    #: one is updated. Those formatoptions are only updated if they exist in
    #: the update parameters.
    children = []

    #: `list of str`. List of formatoptions that force an update of this
    #: formatoption if they are updated.
    dependencies = []

    #: `list of str`. Connections to other formatoptions that are (different
    #: from :attr:`dependencies` and :attr:`children`) not important for the
    #: update process
    connections = []

    #: `list of str`. List of formatoptions that, if included in the update,
    #: prevent the update of this formatoption.
    parents = []

    #: :class:`bool`. Has to be True if the formatoption has a ``make_plot``
    #: method to make the plot.
    plot_fmt = False

    #: :class:`bool`. True if an update of this formatoption requires a
    #: clearing of the axes and reinitialization of the plot
    requires_clearing = False

    #: :class:`str`. Key of the group name in :data:`groups` of this
    #: formatoption keyword
    group = 'misc'

    #: :class:`bool` or a callable. This attribute indicates whether this
    #: :class:`Formatoption` depends on the data and should be updated if the
    #: data changes. If it is a callable, it must accept one argument: the
    #: new data. (Note: This is automatically set to True for plot
    #: formatoptions)
    data_dependent = False

    #: :class:`bool`. True if this formatoption needs an update after the plot
    #: has changed
    update_after_plot = False

    #: :class:`set` of the :class:`Formatoption` instance that are shared
    #: with this instance.
    shared = set()

    #: int or None. Index that is used in case the plotting data is a
    #: :class:`psyplot.InteractiveList`
    index_in_list = 0

    #: :class:`str`. A bit more verbose name than the formatoption key to be
    #: included in the gui. If None, the key is used in the gui
    name = None

    #: Boolean that is True if an update of the formatoption requires a replot
    requires_replot = False

    @property
    def init_kwargs(self):
        """:class:`dict` key word arguments that are passed to the
        initialization of a new instance when accessed from the descriptor"""
        return self._child_mapping

    @property
    def project(self):
        """Project of the plotter of this instance"""
        return self.plotter.project

    @property
    def ax(self):
        """The axes this Formatoption plots on"""
        return self.plotter.ax

    @property
    def lock(self):
        """A :class:`threading.Rlock` instance to lock while updating

        This lock is used when multiple :class:`plotter` instances are
        updated at the same time while sharing formatoptions."""
        try:
            return self._lock
        except AttributeError:
            self._lock = RLock()
            return self._lock

    @property
    def logger(self):
        """Logger of the plotter"""
        return self.plotter.logger.getChild(self.key)

    @property
    def groupname(self):
        """Long name of the group this formatoption belongs too."""
        try:
            return groups[self.group]
        except KeyError:
            warn("Unknown formatoption group " + str(self.group),
                 PsyPlotRuntimeWarning)
            return self.group

    @property
    def raw_data(self):
        """The original data of the plotter of this formatoption"""
        if self.index_in_list is not None and isinstance(
                self.plotter.data, InteractiveList):
            return self.plotter.data[self.index_in_list]
        else:
            return self.plotter.data

    @property
    def decoder(self):
        """The :class:`~psyplot.data.CFDecoder` instance that decodes the
        :attr:`raw_data`"""
        # If the decoder is modified by one of the formatoptions, use this one
        if self.plotter.plot_data_decoder is not None:
            if self.index_in_list is not None and isinstance(
                    self.plotter.plot_data, InteractiveList):
                ret = self.plotter.plot_data_decoder[self.index_in_list]
                if ret is not None:
                    return ret
            else:
                return self.plotter.plot_data_decoder
        data = self.raw_data
        check = isinstance(data, InteractiveList)
        while check:
            data = data[0]
            check = isinstance(data, InteractiveList)
        return data.psy.decoder

    @decoder.setter
    def decoder(self, value):
        self.set_decoder(value, self.index_in_list)

    @property
    def any_decoder(self):
        """Return the first possible decoder"""
        ret = self.decoder
        while not isinstance(ret, CFDecoder):
            ret = ret[0]
        return ret

    @property
    def data(self):
        """The data that is plotted"""
        if self.index_in_list is not None and isinstance(
                self.plotter.plot_data, InteractiveList):
            return self.plotter.plot_data[self.index_in_list]
        else:
            return self.plotter.plot_data

    @data.setter
    def data(self, value):
        self.set_data(value, self.index_in_list)

    @property
    def iter_data(self):
        """Returns an iterator over the plot data arrays"""
        data = self.data
        if isinstance(data, InteractiveList):
            return iter(data)
        return iter([data])

    @property
    def iter_raw_data(self):
        """Returns an iterator over the original data arrays"""
        data = self.raw_data
        if isinstance(data, InteractiveList):
            return iter(data)
        return iter([data])

    @property
    def validate(self):
        """Validation method of the formatoption"""
        try:
            return self._validate
        except AttributeError:
            try:
                self._validate = self.plotter.get_vfunc(self.key)
            except KeyError:
                warn("Could not find a validation function for %s "
                     "formatoption keyword! No validation will be made!" % (
                         self.key), PsyPlotRuntimeWarning, logger=self.logger)
                self._validate = _identity
        return self._validate

    @validate.setter
    def validate(self, value):
        self._validate = value

    @property
    def default(self):
        """Default value of this formatoption"""
        return self.plotter.rc[self.key]

    @property
    def default_key(self):
        """The key of this formatoption in the :attr:`psyplot.rcParams`"""
        return self.plotter.rc._get_val_and_base(self.key)[0]

    @property
    def shared_by(self):
        """None if the formatoption is not controlled by another formatoption
        of another plotter, otherwise the corresponding :class:`Formatoption`
        instance"""
        return self.plotter._shared.get(self.key)

    @property
    def value(self):
        """Value of the formatoption in the corresponding :attr:`plotter` or
        the shared value"""
        shared_by = self.shared_by
        if shared_by:
            return shared_by.value2share
        return self.plotter[self.key]

    @property
    @dedent
    def changed(self):
        """
        :class:`bool` indicating whether the value changed compared to the
        default or not."""
        return self.diff(self.default)

    @property
    @dedent
    def value2share(self):
        """
        The value that is passed to shared formatoptions (by default, the
        :attr:`value` attribute)"""
        return self.value

    @property
    @dedent
    def value2pickle(self):
        """
        The value that can be used when pickling the information of the project
        """
        return self.value

    @docstrings.get_sections(base='Formatoption')
    @dedent
    def __init__(self, key, plotter=None, index_in_list=None,
                 additional_children=[], additional_dependencies=[],
                 **kwargs):
        """
        Parameters
        ----------
        key: str
            formatoption key in the `plotter`
        plotter: psyplot.plotter.Plotter
            Plotter instance that holds this formatoption. If None, it is
            assumed that this instance serves as a descriptor.
        index_in_list: int or None
            The index that shall be used if the data is a
            :class:`psyplot.InteractiveList`
        additional_children: list or str
            Additional children to use (see the :attr:`children` attribute)
        additional_dependencies: list or str
            Additional dependencies to use (see the :attr:`dependencies`
            attribute)
        ``**kwargs``
            Further keywords may be used to specify different names for
            children, dependencies and connection formatoptions that match the
            setup of the plotter. Hence, keywords may be anything of the
            :attr:`children`, :attr:`dependencies` and :attr:`connections`
            attributes, with values being the name of the new formatoption in
            this plotter."""
        self.key = key
        self.plotter = plotter
        self.index_in_list = index_in_list
        self.shared = set()
        self.additional_children = additional_children
        self.additional_dependencies = additional_dependencies
        self.children = self.children + additional_children
        self.dependencies = self.dependencies + additional_dependencies
        self._child_mapping = dict(zip(*tee(chain(
            self.children, self.dependencies, self.connections,
            self.parents), 2)))
        # check kwargs
        for key in (key for key in kwargs if key not in self._child_mapping):
            raise TypeError(
                '%s.__init__() got an unexpected keyword argument %r' % (
                    self.__class__.__name__, key))
        # set up child mapping
        self._child_mapping.update(kwargs)
        # reset the dependency lists to match the current plotter setup
        for attr in ['children', 'dependencies', 'connections', 'parents']:
            setattr(self, attr,
                    [self._child_mapping[key] for key in getattr(self, attr)])

    def __set__(self, instance, value):
        if isinstance(value, Formatoption):
            setattr(instance, '_' + self.key, value)
        else:
            fmto = getattr(instance, self.key)
            fmto.set_value(value)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        try:
            return getattr(instance, '_' + self.key)
        except AttributeError:
            fmto = self.__class__(
                self.key, instance, self.index_in_list,
                additional_children=self.additional_children,
                additional_dependencies=self.additional_dependencies,
                **self.init_kwargs)
            setattr(instance, '_' + self.key, fmto)
            return fmto

    def __delete__(self, instance, owner):
        fmto = getattr(instance, '_' + self.key)
        with instance.no_validation:
            instance[self.key] = fmto.default

    @docstrings.get_sections(base='Formatoption.set_value')
    @dedent
    def set_value(self, value, validate=True, todefault=False):
        """
        Set (and validate) the value in the plotter. This method is called by
        the plotter when it attempts to change the value of the formatoption.

        Parameters
        ----------
        value
            Value to set
        validate: bool
            if True, validate the `value` before it is set
        todefault: bool
            True if the value is updated to the default value"""
        # do nothing if the key is shared
        if self.key in self.plotter._shared:
            return
        with self.plotter.no_validation:
            self.plotter[self.key] = value if not validate else \
                self.validate(value)

    def set_data(self, data, i=None):
        """
        Replace the data to plot

        This method may be used to replace the data that is visualized by the
        plotter. It changes it's behaviour depending on whether an
        :class:`psyplot.data.InteractiveList` is visualized or a single
        :class:`pysplot.data.InteractiveArray`

        Parameters
        ----------
        data: psyplot.data.InteractiveBase
            The data to insert
        i: int
            The position in the InteractiveList where to insert the data (if
            the plotter visualizes a list anyway)

        Notes
        -----
        This method uses the :attr:`Formatoption.data` attribute
        """
        if self.index_in_list is not None:
            i = self.index_in_list
        if i is not None and isinstance(self.plotter.plot_data,
                                        InteractiveList):
            self.plotter.plot_data[i] = data
        else:
            self.plotter.plot_data = data

    def set_decoder(self, decoder, i=None):
        """
        Replace the data to plot

        This method may be used to replace the data that is visualized by the
        plotter. It changes it's behaviour depending on whether an
        :class:`psyplot.data.InteractiveList` is visualized or a single
        :class:`pysplot.data.InteractiveArray`

        Parameters
        ----------
        decoder: psyplot.data.CFDecoder
            The decoder to insert
        i: int
            The position in the InteractiveList where to insert the data (if
            the plotter visualizes a list anyway)
        """
        # we do not modify the raw data but instead set it on the plotter
        # TODO: This is not safe for encapsulated InteractiveList instances!
        if i is not None and isinstance(
                self.plotter.plot_data, InteractiveList):
            n = len(self.plotter.plot_data)
            decoders = self.plotter.plot_data_decoder or [None] * n
            decoders[i] = decoder
            self.plotter.plot_data_decoder = decoders
        else:
            if (isinstance(self.plotter.plot_data, InteractiveList) and
                    isinstance(decoder, CFDecoder)):
                decoder = [decoder] * len(self.plotter.plot_data)
            self.plotter.plot_data_decoder = decoder

    def get_decoder(self, i=None):
        # we do not modify the raw data but instead set it on the plotter
        # TODO: This is not safe for encapsulated InteractiveList instances!
        if i is not None and isinstance(
                self.plotter.plot_data, InteractiveList):
            n = len(self.plotter.plot_data)
            decoders = self.plotter.plot_data_decoder or [None] * n
            return decoders[i] or self.plotter.plot_data[i].psy.decoder
        else:
            return self.decoder

    def check_and_set(self, value, todefault=False, validate=True):
        """Checks the value and sets the value if it changed

        This method checks the value and sets it only if the :meth:`diff`
        method result of the given `value` is True

        Parameters
        ----------
        value
            A possible value to set
        todefault: bool
            True if the value is updated to the default value

        Returns
        -------
        bool
            A boolean to indicate whether it has been set or not"""
        if validate:
            value = self.validate(value)
        if self.diff(value):
            self.set_value(value, validate=False, todefault=todefault)
            return True
        return False

    def diff(self, value):
        """Checks whether the given value differs from what is currently set

        Parameters
        ----------
        value
            A possible value to set (make sure that it has been validate via
            the :attr:`validate` attribute before)

        Returns
        -------
        bool
            True if the value differs from what is currently set"""
        return value != self.value

    def initialize_plot(self, value, *args, **kwargs):
        """Method that is called when the plot is made the first time

        Parameters
        ----------
        value
            The value to use for the initialization"""
        self.update(value, *args, **kwargs)

    @abstractmethod
    def update(self, value):
        """Method that is call to update the formatoption on the axes

        Parameters
        ----------
        value
            Value to update"""
        pass

    def get_fmt_widget(self, parent, project):
        """Get a widget to update the formatoption in the GUI

        This method should return a QWidget that is loaded by the psyplot-gui
        when the formatoption is selected in the
        :attr:`psyplot_gui.main.Mainwindow.fmt_widget`. It should call the
        :meth:`~psyplot_gui.fmt_widget.FormatoptionWidget.insert_text` method
        when the update text for the formatoption should be changed.

        Parameters
        ----------
        parent: psyplot_gui.fmt_widget.FormatoptionWidget
            The parent widget that contains the returned QWidget
        project: psyplot.project.Project
            The current subproject (see :func:`psyplot.project.gcp`)

        Returns
        -------
        PyQt5.QtWidgets.QWidget
            The widget to control the formatoption"""
        return None

    def share(self, fmto, initializing=False, **kwargs):
        """Share the settings of this formatoption with other data objects

        Parameters
        ----------
        fmto: Formatoption
            The :class:`Formatoption` instance to share the attributes with
        ``**kwargs``
            Any other keyword argument that shall be passed to the update
            method of `fmto`"""
        # lock all  the childrens and the formatoption itself
        self.lock.acquire()
        fmto._lock_children()
        fmto.lock.acquire()
        # update the other plotter
        if initializing:
            fmto.initialize_plot(self.value2share, **kwargs)
        else:
            fmto.update(self.value2share, **kwargs)
        self.shared.add(fmto)
        # release the locks
        fmto.lock.release()
        fmto._release_children()
        self.lock.release()

    def _lock_children(self):
        """acquire the locks of the children"""
        plotter = self.plotter
        for key in self.children + self.dependencies:
            try:
                getattr(plotter, key).lock.acquire()
            except AttributeError:
                pass

    def _release_children(self):
        """release the locks of the children"""
        plotter = self.plotter
        for key in self.children + self.dependencies:
            try:
                getattr(plotter, key).lock.release()
            except AttributeError:
                pass

    def finish_update(self):
        """Finish the update, initialization and sharing process

        This function is called at the end of the :meth:`Plotter.start_update`,
        :meth:`Plotter.initialize_plot` or the :meth:`Plotter.share` methods.
        """
        pass

    @dedent
    def remove(self):
        """
        Method to remove the effects of this formatoption

        This method is called when the axes is cleared due to a
        formatoption with :attr:`requires_clearing` set to True. You don't
        necessarily have to implement this formatoption if your plot results
        are removed by the usual :meth:`matplotlib.axes.Axes.clear` method."""
        pass


class DictFormatoption(Formatoption):
    """
    Base formatoption class defining an alternative set_value that works for
    dictionaries."""

    @docstrings.dedent
    def set_value(self, value, validate=True, todefault=False):
        """
        Set (and validate) the value in the plotter

        Parameters
        ----------
        %(Formatoption.set_value.parameters)s

        Notes
        -----
        - If the current value in the plotter is None, then it will be set with
          the given `value`, otherwise the current value in the plotter is
          updated
        - If the value is an empty dictionary, the value in the plotter is
          cleared"""
        value = value if not validate else self.validate(value)
        # if the key in the plotter is not already set (i.e. it is initialized
        # with None, we set it)
        if self.plotter[self.key] is None:
            with self.plotter.no_validation:
                self.plotter[self.key] = value.copy()
        # in case of an empty dict, clear the value
        elif not value:
            self.plotter[self.key].clear()
        # otherwhise we update the dictionary
        else:
            if todefault:
                self.plotter[self.key].clear()
            self.plotter[self.key].update(value)


class PostTiming(Formatoption):
    """
    Determine when to run the :attr:`post` formatoption

    This formatoption determines, whether the :attr:`post` formatoption
    should be run never, after replot or after every update.

    Possible types
    --------------
    'never'
        Never run post processing scripts
    'always'
        Always run post processing scripts
    'replot'
        Only run post processing scripts when the data changes or a replot
        is necessary

    See Also
    --------
    post: The post processing formatoption"""

    default = 'never'

    priority = -inf

    group = 'post_processing'

    name = 'Timing of the post processing'

    @staticmethod
    def validate(value):
        value = six.text_type(value)
        possible_values = ['never', 'always', 'replot']
        if value not in possible_values:
            raise ValueError('String must be one of %s, not %r' % (
                possible_values, value))
        return value

    def update(self, value):
        pass

    def get_fmt_widget(self, parent, project):
        from psyplot_gui.compat.qtcompat import QComboBox
        combo = QComboBox(parent)
        combo.addItems(['never', 'always', 'replot'])
        combo.setCurrentText(
            next((plotter[self.key] for plotter in project.plotters), 'never'))
        combo.currentTextChanged.connect(parent.set_obj)
        return combo


class PostProcDependencies(object):
    """The dependencies of this formatoption"""

    def __get__(self, instance, owner):
        if (instance is None or instance.plotter is None or
                not instance.plotter._initialized):
            return []
        elif instance.post_timing.value == 'always':
            return list(set(instance.plotter) - {instance.key})
        else:
            return []

    def __set__(self, instance, value):
        pass


class PostProcessing(Formatoption):
    """
    Apply your own postprocessing script

    This formatoption let's you apply your own post processing script. Just
    enter the script as a string and it will be executed. The formatoption
    will be made available via the ``self`` variable

    Possible types
    --------------
    None
        Don't do anything
    str
        The post processing script as string

    Note
    ----
    This formatoption uses the built-in :func:`exec` function to compile the
    script. Since this poses a security risk when loading psyplot projects,
    it is by default disabled through the :attr:`Plotter.enable_post`
    attribute. If you are sure that you can trust the script in this
    formatoption, set this attribute of the corresponding :class:`Plotter` to
    ``True``

    Examples
    --------
    Assume, you want to manually add the mean of the data to the title of the
    matplotlib axes. You can simply do this via

    .. code-block:: python

        from psyplot.plotter import Plotter
        from xarray import DataArray
        plotter = Plotter(DataArray([1, 2, 3]))
        # enable the post formatoption
        plotter.enable_post = True
        plotter.update(post="self.ax.set_title(str(self.data.mean()))")
        plotter.ax.get_title()
        '2.0'

    By default, the ``post`` formatoption is only ran, when it is explicitly
    updated. However, you can use the :attr:`post_timing` formatoption, to
    run it automatically. E.g. for running it after every update of the
    plotter, you can set

    .. code-block:: python

        plotter.update(post_timing='always')

    See Also
    --------
    post_timing: Determine the timing of this formatoption"""

    children = ['post_timing']

    default = None

    priority = -inf

    group = 'post_processing'

    name = 'Custom post processing script'

    @staticmethod
    def validate(value):
        if value is None:
            return value
        elif not isinstance(value, six.string_types):
            raise ValueError("Expected a string, not %s" % (type(value), ))
        else:
            return six.text_type(value)

    @property
    def data_dependent(self):
        """True if the corresponding :class:`post_timing <PostTiming>`
        formatoption is set to ``'replot'`` to run the post processing script
        after every change of the data"""
        return self.post_timing.value == 'replot'

    dependencies = PostProcDependencies()

    def update(self, value):
        if value is None:
            return
        if not self.plotter.enable_post:
            warn(
                "Post processing is disabled. Set the ``enable_post`` "
                "attribute to True to run the script")
        else:
            exec(value, {'self': self})


class Plotter(dict):
    """Interactive plotting object for one or more data arrays

    This class is the base for the interactive plotting with the psyplot
    module. It capabilities are determined by it's descriptor classes that are
    derived from the :class:`Formatoption` class"""

    #: List of base strings in the :attr:`psyplot.rcParams` dictionary
    _rcparams_string = []

    post_timing = PostTiming('post_timing')
    post = PostProcessing('post')

    no_validation = _temp_bool_prop('no_validation', """
        Temporarily disable the validation

        Examples
        --------
        Although it is not recommended to set a value with disabled validation,
        you can disable it via::

            >>> with plotter.no_validation:
            ...     plotter['ticksize'] = 'x'

        To permanently disable the validation, simply set

            >>> plotter.no_validation = True
            >>> plotter['ticksize'] = 'x'
            >>> plotter.no_validation = False  # reenable validation""")

    #: Temporarily include links in the key descriptions from
    #: :meth:`show_keys`, :meth:`show_docs` and :meth:`show_summaries`.
    #: Note that this is a class attribute, so each change to the value of this
    #: attribute will affect all instances and subclasses
    include_links = _TempBool()

    @property
    def ax(self):
        """Axes instance of the plot"""
        if self._ax is None:
            import matplotlib.pyplot as plt
            plt.figure()
            self._ax = plt.axes(projection=self._get_sample_projection())
        return self._ax

    @ax.setter
    def ax(self, value):
        self._ax = value

    #: The :class:`psyplot.project.Project` instance this plotter belongs to
    _project = None

    @property
    def project(self):
        """:class:`psyplot.project.Project` instance this plotter belongs to"""
        if self._project is None:
            return
        return self._project()

    @project.setter
    def project(self, value):
        if value is not None:
            self._project = weakref.ref(value)
        else:
            self._project = value

    @property
    @dedent
    def rc(self):
        """
        Default values for this plotter

        This :class:`~psyplot.config.rcsetup.SubDict` stores the default values
        for this plotter. A modification of the dictionary does not affect
        other plotter instances unless you set the
        :attr:`~psyplot.config.rcsetup.SubDict.trace` attribute to True"""
        try:
            return self._rc
        except AttributeError:
            self._set_rc()
            return self._rc

    @property
    def base_variables(self):
        """A mapping from the base_variable names to the variables"""
        if isinstance(self.data, InteractiveList):
            return dict(chain(*map(
                lambda arr: six.iteritems(arr.psy.base_variables),
                self.data)))
        else:
            return self.data.psy.base_variables

    @property
    def iter_base_variables(self):
        """A mapping from the base_variable names to the variables"""
        if isinstance(self.data, InteractiveList):
            return chain(*(arr.psy.iter_base_variables for arr in self.data))
        else:
            return self.data.psy.iter_base_variables

    no_auto_update = property(_no_auto_update_getter,
                              doc=_no_auto_update_getter.__doc__)

    @no_auto_update.setter
    def no_auto_update(self, value):
        self.no_auto_update.value = bool(value)

    @property
    def changed(self):
        """:class:`dict` containing the key value pairs that are not the
        default"""
        return {key: value for key, value in six.iteritems(self)
                if getattr(self, key).changed}

    @property
    def figs2draw(self):
        """All figures that have been manipulated through sharing and the own
        figure.

        Notes
        -----
        Using this property set will reset the figures too draw"""
        return self._figs2draw.union([self.ax.get_figure()])

    @property
    @docstrings
    def _njobs(self):
        """%(InteractiveBase._njobs)s"""
        if self.disabled:
            return [0]
        return [1, 1]

    @property
    def _fmtos(self):
        """Iterator over the formatoptions"""
        return (getattr(self, key) for key in self)

    @property
    def _fmto_groups(self):
        """Mapping from group to a set of formatoptions"""
        ret = defaultdict(set)
        for key in self:
            ret[getattr(self, key).group].add(getattr(self, key))
        return dict(ret)

    @property
    def fmt_groups(self):
        """A mapping from the formatoption group to the formatoptions"""
        ret = defaultdict(set)
        for key in self:
            ret[getattr(self, key).group].add(key)
        return dict(ret)

    @property
    def groups(self):
        """A mapping from the group short name to the group description"""
        return {group: groups[group] for group in self.fmt_groups}

    @property
    def data(self):
        """The :class:`psyplot.InteractiveBase` instance of this plotter"""
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def plot_data(self):
        """The data that is used for plotting"""
        return getattr(self, '_plot_data', self.data)

    @plot_data.setter
    def plot_data(self, value):
        self._set_data(value)

    #: The decoder to use for the formatoptions. If None, the decoder of the
    #: raw data is used
    plot_data_decoder = None

    #: :class:`bool` that has to be ``True`` if the post processing script in
    #: the :attr:`post` formatoption should be enabled
    enable_post = False

    def _set_data(self, value):
        if isinstance(value, InteractiveList):
            self._plot_data = value.copy()
        else:
            self._plot_data = value

    @property
    def logger(self):
        """:class:`logging.Logger` of this plotter"""
        try:
            return self.data.psy.logger.getChild(self.__class__.__name__)
        except AttributeError:
            name = '%s.%s' % (self.__module__, self.__class__.__name__)
            return logging.getLogger(name)

    docstrings.keep_params('InteractiveBase.parameters', 'auto_update')

    @docstrings.get_sections(base='Plotter')
    @docstrings.dedent
    def __init__(self, data=None, ax=None, auto_update=None, project=None,
                 draw=False, make_plot=True, clear=False,
                 enable_post=False, **kwargs):
        """
        Parameters
        ----------
        data: InteractiveArray or ArrayList, optional
            Data object that shall be visualized. If given and `plot` is True,
            the :meth:`initialize_plot` method is called at the end. Otherwise
            you can call this method later by yourself
        ax: matplotlib.axes.Axes
            Matplotlib Axes to plot on. If None, a new one will be created as
            soon as the :meth:`initialize_plot` method is called
        %(InteractiveBase.parameters.auto_update)s
        %(InteractiveBase.start_update.parameters.draw)s
        make_plot: bool
            If True, and `data` is not None, the plot is initialized. Otherwise
            only the framework between plotter and data is set up
        clear: bool
            If True, the axes is cleared first
        enable_post: bool
            If True, the :attr:`post` formatoption is enabled and post
            processing scripts are allowed
        ``**kwargs``
            Any formatoption key from the :attr:`formatoptions` attribute that
            shall be used"""
        self.project = project
        self.ax = ax
        self.data = data
        self.enable_post = enable_post
        if auto_update is None:
            auto_update = rcParams['lists.auto_update']
        self.no_auto_update = not bool(auto_update)
        self._registered_updates = {}
        self._todefault = False
        self._old_fmt = []
        self._figs2draw = set()
        #: formatoptions that have to be updated by other plotters that share
        #: the given formatoption with this Plotter. :attr:`_to_update` is a
        #: mapping from the formatoptions in this plotter to the corresponding
        #: other plotter
        self._to_update = {}
        self.disabled = False
        #: Dictionary holding the Formatoption instances of other plotters
        #: if their value shall be used instead of the one in this instance
        self._shared = {}
        #: list of str. Formatoption keys that were changed during the last
        #: update
        self._last_update = []
        #: The set of formatoptions that shall be updated even if they did not
        #: change
        self._force = set()
        self.replot = True
        self.cleared = clear
        self._updating = False
        # will be set to True when the plot is first initialized
        self._initialized = False

        # first we initialize all keys with None. This is necessary in order
        # to make the validation functioning
        with self.no_validation:
            for key in self._get_formatoptions():
                self[key] = None
        for key in self:  # then we set the default values
            fmto = getattr(self, key)
            self._try2set(fmto, fmto.default, validate=False)
        self._set_rc()
        for key, value in six.iteritems(kwargs):  # then the user values
            self[key] = value
        self.initialize_plot(data, ax=ax, draw=draw, clear=clear,
                             make_plot=make_plot)

    def _try2set(self, fmto, *args, **kwargs):
        """Sets the value in `fmto` and gives additional informations when fail

        Parameters
        ----------
        fmto: Formatoption
        ``*args`` and ``**kwargs``
            Anything that is passed to `fmto`s :meth:`~Formatoption.set_value`
            method"""
        fmto.set_value(*args, **kwargs)

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            self.check_key(key)

    def __setitem__(self, key, value):
        if not self.no_validation:
            self.check_key(key)
            self._try2set(getattr(self, key), value)
            return
        # prevent from setting during an update process
        getattr(self, key).lock.acquire()
        dict.__setitem__(self, key, value)
        getattr(self, key).lock.release()

    def __delitem__(self, key):
        self[key] = getattr(self, key).default

    docstrings.delete_params('check_key.parameters', 'possible_keys', 'name')

    @docstrings.dedent
    def check_key(self, key, raise_error=True, *args, **kwargs):
        """
        Checks whether the key is a valid formatoption

        Parameters
        ----------
        %(check_key.parameters.no_possible_keys|name)s

        Returns
        -------
        %(check_key.returns)s

        Raises
        ------
        %(check_key.raises)s"""
        return check_key(
            key, possible_keys=list(self), raise_error=raise_error,
            name='formatoption keyword', *args, **kwargs)

    @classmethod
    @docstrings.get_sections(base='Plotter.check_data', sections=['Parameters',
                                                              'Returns'])
    @dedent
    def check_data(cls, name, dims, is_unstructured):
        """
        A validation method for the data shape

        The default method does nothing and should be subclassed to validate
        the results. If the plotter accepts a :class:`InteractiveList`, it
        should accept a list for name and dims

        Parameters
        ----------
        name: str or list of str
            The variable name(s) of the data
        dims: list of str or list of lists of str
            The dimension name(s) of the data
        is_unstructured: bool or list of bool
            True if the corresponding array is unstructured

        Returns
        -------
        list of bool or None
            True, if everything is okay, False in case of a serious error,
            None if it is intermediate. Each object in this list corresponds to
            one in the given `name`
        list of str
            The message giving more information on the reason. Each object in
            this list corresponds to one in the given `name`"""
        if isinstance(name, six.string_types):
            name = [name]
            dims = [dims]
            is_unstructured = [is_unstructured]
        N = len(name)
        if len(dims) != N or len(is_unstructured) != N:
            return [False] * N, [
                'Number of provided names (%i) and dimensions '
                '(%i) or unstructured information (%i) are not the same' % (
                    N, len(dims), len(is_unstructured))] * N
        return [True] * N, [''] * N

    docstrings.keep_params('Plotter.parameters', 'ax', 'make_plot', 'clear')

    @docstrings.dedent
    def initialize_plot(self, data=None, ax=None, make_plot=True, clear=False,
                        draw=False, remove=False, priority=None):
        """
        Initialize the plot for a data array

        Parameters
        ----------
        data: InteractiveArray or ArrayList, optional
            Data object that shall be visualized.

            - If not None and `plot` is True, the given data is visualized.
            - If None and the :attr:`data` attribute is not None, the data in
              the :attr:`data` attribute is visualized
            - If both are None, nothing is done.
        %(Plotter.parameters.ax|make_plot|clear)s
        %(InteractiveBase.start_update.parameters.draw)s
        remove: bool
            If True, old effects by the formatoptions in this plotter are
            undone first
        priority: int
            If given, initialize only the formatoption with the given priority.
            This value must be out of :data:`START`, :data:`BEFOREPLOTTING` or
            :data:`END`
        """
        if data is None and self.data is not None:
            data = self.data
        else:
            self.data = data
        self.ax = ax
        if data is None:  # nothing to do if no data is given
            return
        self.no_auto_update = not (
            not self.no_auto_update or not data.psy.no_auto_update)
        data.psy.plotter = self
        if not make_plot:  # stop here if we shall not plot
            return
        self.logger.debug("Initializing plot...")
        if remove:
            self.logger.debug("    Removing old formatoptions...")
            for fmto in self._fmtos:
                try:
                    fmto.remove()
                except Exception:
                    self.logger.debug(
                        "Could not remove %s while initializing", fmto.key,
                        exc_info=True)
        if clear:
            self.logger.debug("    Clearing axes...")
            self.ax.clear()
            self.cleared = True
        # get the formatoptions. We sort them here by key to make sure that the
        # order always stays the same (easier for debugging)
        fmto_groups = self._grouped_fmtos(self._sorted_by_priority(
            sorted(self._fmtos, key=lambda fmto: fmto.key)))
        self.plot_data = self.data
        self._updating = True
        for fmto_priority, grouper in fmto_groups:
            if priority is None or fmto_priority == priority:
                self._plot_by_priority(fmto_priority, grouper,
                                       initializing=True)
        self._release_all(True)  # finish the update
        self.cleared = False
        self.replot = False
        self._initialized = True
        self._updating = False

        if draw is None:
            draw = rcParams['auto_draw']
        if draw:
            self.draw()
            if rcParams['auto_show']:
                self.show()

    docstrings.keep_params('InteractiveBase._register_update.parameters',
                           'force', 'todefault')

    @docstrings.get_sections(base='Plotter._register_update')
    @docstrings.dedent
    def _register_update(self, fmt={}, replot=False, force=False,
                         todefault=False):
        """
        Register formatoptions for the update

        Parameters
        ----------
        fmt: dict
            Keys can be any valid formatoptions with the corresponding values
            (see the :attr:`formatoptions` attribute)
        replot: bool
            Boolean that determines whether the data specific formatoptions
            shall be updated in any case or not.
        %(InteractiveBase._register_update.parameters.force|todefault)s"""
        if self.disabled:
            return
        self.replot = self.replot or replot
        self._todefault = self._todefault or todefault
        if force is True:
            force = list(fmt)
        self._force.update(
            [ret[0] for ret in map(self.check_key, force or [])])
        # check the keys
        list(map(self.check_key, fmt))
        self._registered_updates.update(fmt)

    def make_plot(self):
        """Method for making the plot

        This method is called at the end of the :attr:`BEFOREPLOTTING` stage if
        and only if the :attr:`plot_fmt` attribute is set to ``True``"""
        pass

    @docstrings.dedent
    def start_update(self, draw=None, queues=None, update_shared=True):
        """
        Conduct the registered plot updates

        This method starts the updates from what has been registered by the
        :meth:`update` method. You can call this method if you did not set the
        `auto_update` parameter to True when calling the :meth:`update` method
        and when the :attr:`no_auto_update` attribute is True.

        Parameters
        ----------
        %(InteractiveBase.start_update.parameters)s

        Returns
        -------
        %(InteractiveBase.start_update.returns)s

        See Also
        --------
        :attr:`no_auto_update`, update"""
        def update_the_others():
            for fmto in fmtos:
                for other_fmto in fmto.shared:
                    if not other_fmto.plotter._updating:
                        other_fmto.plotter._register_update(
                            force=[other_fmto.key])
            for fmto in fmtos:
                for other_fmto in fmto.shared:
                    if not other_fmto.plotter._updating:
                        other_draw = other_fmto.plotter.start_update(
                            draw=False, update_shared=False)
                        if other_draw:
                            self._figs2draw.add(
                                other_fmto.plotter.ax.get_figure())
        if self.disabled:
            return False

        if queues is not None:
            queues[0].get()
        self.logger.debug("Starting update of %r",
                          self._registered_updates.keys())
        # update the formatoptions
        self._save_state()
        try:
            # get the formatoptions. We sort them here by key to make sure that
            # the order always stays the same (easier for debugging)
            fmtos = sorted(self._set_and_filter(), key=lambda fmto: fmto.key)
        except Exception:
            # restore last (working) state
            last_state = self._old_fmt.pop(-1)
            with self.no_validation:
                for key in self:
                    self[key] = last_state.get(key, getattr(self, key).default)
            if queues is not None:
                queues[0].task_done()
            self._release_all(queue=None if queues is None else queues[1])
            # raise the error
            raise
        for fmto in fmtos:
            for fmto2 in fmto.shared:
                fmto2.plotter._to_update[fmto2] = self
        if queues is not None:
            self._updating = True
            queues[0].task_done()
            # wait for the other tasks to finish
            queues[0].join()
            queues[1].get()
        fmtos.extend([fmto for fmto in self._insert_additionals(list(
            self._to_update)) if fmto not in fmtos])
        self._to_update.clear()

        fmto_groups = self._grouped_fmtos(self._sorted_by_priority(fmtos[:]))
        # if any formatoption requires a clearing of the axes is updated,
        # we reinitialize the plot
        try:
            if self.cleared:
                self.reinit(draw=draw)
                update_the_others()
                arr_draw = True
            else:
                # otherwise we update it
                arr_draw = False
                for priority, grouper in fmto_groups:
                    arr_draw = True
                    self._plot_by_priority(priority, grouper)
                update_the_others()
        except Exception:
            raise
        finally:
            # make sure that all locks are released
            self._release_all(finish=True,
                              queue=None if queues is None else queues[1])
        if draw is None:
            draw = rcParams['auto_draw']
        if draw and arr_draw:
            self.draw()
            if rcParams['auto_show']:
                self.show()
        self.replot = False
        return arr_draw

    def _release_all(self, finish=False, queue=None):
        # make sure that all locks are released
        try:
            for fmto in self._fmtos:
                if finish:
                    fmto.finish_update()
                try:
                    fmto.lock.release()
                except RuntimeError:
                    pass
        except:
            raise
        finally:
            if queue is not None:
                queue.task_done()
                queue.join()
            self._updating = False

    def _plot_by_priority(self, priority, fmtos, initializing=False):
        def update(fmto):
            other_fmto = self._shared.get(fmto.key)
            if other_fmto:
                self.logger.debug("%s is shared with %s", fmto.key,
                                  other_fmto.plotter.logger.name)
                other_fmto.share(fmto, initializing=initializing)
            # but if not, share them
            else:
                if initializing:
                    self.logger.debug("Initializing %s", fmto.key)
                    fmto.initialize_plot(fmto.value)
                else:
                    self.logger.debug("Updating %s", fmto.key)
                    fmto.update(fmto.value)
            try:
                fmto.lock.release()
            except RuntimeError:
                pass

        self._initializing = initializing

        self.logger.debug(
            "%s formatoptions with priority %i",
            "Initializing" if initializing else "Updating", priority)

        if priority >= START or priority == END:
            for fmto in fmtos:
                update(fmto)
        elif priority == BEFOREPLOTTING:
            for fmto in fmtos:
                update(fmto)
            self._make_plot()

        self._initializing = False

    @docstrings.dedent
    def reinit(self, draw=None, clear=False):
        """
        Reinitializes the plot with the same data and on the same axes.

        Parameters
        ----------
        %(InteractiveBase.start_update.parameters.draw)s
        clear: bool
            Whether to clear the axes or not

        Warnings
        --------
        The axes may be cleared when calling this method (even if `clear` is
        set to False)!"""
        # call the initialize_plot method. Note that clear can be set to
        # False if any fmto has requires_clearing attribute set to True,
        # because this then has been cleared before
        self.initialize_plot(
            self.data, self._ax, draw=draw, clear=clear or any(
                fmto.requires_clearing for fmto in self._fmtos),
            remove=True)

    def draw(self):
        """Draw the figures and those that are shared and have been changed"""
        for fig in self.figs2draw:
            fig.canvas.draw()
        self._figs2draw.clear()

    def _grouped_fmtos(self, fmtos):
        def key_func(fmto):
            if fmto.priority >= START:
                return START
            elif fmto.priority >= BEFOREPLOTTING:
                return BEFOREPLOTTING
            else:
                return END
        return groupby(fmtos, key_func)

    def _set_and_filter(self):
        """Filters the registered updates and sort out what is not needed

        This method filters out the formatoptions that have not changed, sets
        the new value and returns an iterable that is sorted by the priority
        (highest priority comes first) and dependencies

        Returns
        -------
        list
            list of :class:`Formatoption` objects that have to be updated"""
        fmtos = []
        seen = set()
        for key in self._force:
            self._registered_updates.setdefault(key, getattr(self, key).value)
        for key, value in chain(
                six.iteritems(self._registered_updates),
                six.iteritems(
                    {key: getattr(self, key).default for key in self})
                if self._todefault else ()):
            if key in seen:
                continue
            seen.add(key)
            fmto = getattr(self, key)
            # if the key is shared, a warning will be printed as long as
            # this plotter is not also updating (for example due to a whole
            # project update)
            if key in self._shared and key not in self._force:
                if not self._shared[key].plotter._updating:
                    warn(("%s formatoption is shared with another plotter."
                          " Use the unshare method to enable the updating") % (
                              fmto.key),
                         logger=self.logger)
                changed = False
            else:
                try:
                    changed = fmto.check_and_set(
                        value, todefault=self._todefault,
                        validate=not self.no_validation)
                except Exception as e:
                    self._registered_updates.pop(key, None)
                    self.logger.debug('Failed to set %s', key)
                    raise e
            changed = changed or key in self._force
            if changed:
                fmtos.append(fmto)
        fmtos = self._insert_additionals(fmtos, seen)
        for fmto in fmtos:
            fmto.lock.acquire()
        self._todefault = False
        self._registered_updates.clear()
        self._force.clear()
        return fmtos

    def _insert_additionals(self, fmtos, seen=None):
        """
        Insert additional formatoptions into `fmtos`.

        This method inserts those formatoptions into `fmtos` that are required
        because one of the following criteria is fullfilled:

        1. The :attr:`replot` attribute is True
        2. Any formatoption with START priority is in `fmtos`
        3. A dependency of one formatoption is in `fmtos`

        Parameters
        ----------
        fmtos: list
            The list of formatoptions that shall be updated
        seen: set
            The formatoption keys that shall not be included. If None, all
            formatoptions in `fmtos` are used

        Returns
        -------
        fmtos
            The initial `fmtos` plus further formatoptions

        Notes
        -----
        `fmtos` and `seen` are modified in place (except that any formatoption
        in the initial `fmtos` has :attr:`~Formatoption.requires_clearing`
        attribute set to True)"""
        def get_dependencies(fmto):
            if fmto is None:
                return []
            return fmto.dependencies + list(chain(*map(
                lambda key: get_dependencies(getattr(self, key, None)),
                fmto.dependencies)))
        seen = seen or {fmto.key for fmto in fmtos}
        keys = {fmto.key for fmto in fmtos}
        self.replot = self.replot or any(
            fmto.requires_replot for fmto in fmtos)
        if self.replot or any(fmto.priority >= START for fmto in fmtos):
            self.replot = True
            self.plot_data = self.data
            new_fmtos = dict((f.key, f) for f in self._fmtos
                             if ((f not in fmtos and is_data_dependent(
                                 f, self.data))))
            seen.update(new_fmtos)
            keys.update(new_fmtos)
            fmtos += list(new_fmtos.values())

        # insert the formatoptions that have to be updated if the plot is
        # changed
        if any(fmto.priority >= BEFOREPLOTTING for fmto in fmtos):
            new_fmtos = dict((f.key, f) for f in self._fmtos
                             if ((f not in fmtos and f.update_after_plot)))
            fmtos += list(new_fmtos.values())
        for fmto in set(self._fmtos).difference(fmtos):
            all_dependencies = get_dependencies(fmto)
            if keys.intersection(all_dependencies):
                fmtos.append(fmto)
        if any(fmto.requires_clearing for fmto in fmtos):
            self.cleared = True
            return list(self._fmtos)
        return fmtos

    def _sorted_by_priority(self, fmtos, changed=None):
        """Sort the formatoption objects by their priority and dependency

        Parameters
        ----------
        fmtos: list
            list of :class:`Formatoption` instances
        changed: list
            the list of formatoption keys that have changed

        Yields
        ------
        Formatoption
            The next formatoption as it comes by the sorting

        Warnings
        --------
        The list `fmtos` is cleared by this method!"""
        def pop_fmto(key):
            idx = fmtos_keys.index(key)
            del fmtos_keys[idx]
            return fmtos.pop(idx)

        def get_children(fmto, parents_keys):
            all_fmtos = fmtos_keys + parents_keys
            for key in fmto.children + fmto.dependencies:
                if key not in fmtos_keys:
                    continue
                child_fmto = pop_fmto(key)
                for childs_child in get_children(
                        child_fmto, parents_keys + [child_fmto.key]):
                    yield childs_child
                # filter out if parent is in update list
                if (any(key in all_fmtos for key in child_fmto.parents) or
                        fmto.key in child_fmto.parents):
                    continue
                yield child_fmto

        fmtos.sort(key=lambda fmto: fmto.priority, reverse=True)
        fmtos_keys = [fmto.key for fmto in fmtos]
        self._last_update = changed or fmtos_keys[:]
        self.logger.debug("Update the formatoptions %s", fmtos_keys)
        while fmtos:
            del fmtos_keys[0]
            fmto = fmtos.pop(0)
            # first update children
            for child_fmto in get_children(fmto, [fmto.key]):
                yield child_fmto
            # filter out if parent is in update list
            if any(key in fmtos_keys for key in fmto.parents):
                continue
            yield fmto

    @classmethod
    def _get_formatoptions(cls, include_bases=True):
        """
        Iterator over formatoptions

        This class method returns an iterator that contains all the
        formatoption keys that are in this class and that are defined
        in the base classes

        Notes
        -----
        There is absolutely no need to call this method besides the plotter
        initialization, since all formatoptions are in the plotter itself.
        Just type::

            >>> list(plotter)

        to get the formatoptions.

        See Also
        --------
        _format_keys"""
        def base_fmtos(base):
            return filter(
                lambda key: isinstance(getattr(cls, key), Formatoption),
                getattr(base, '_get_formatoptions', empty)(False))

        def empty(*args, **kwargs):
            return list()
        fmtos = (attr for attr, obj in six.iteritems(cls.__dict__)
                 if isinstance(obj, Formatoption))
        if not include_bases:
            return fmtos
        return unique_everseen(chain(fmtos, *map(base_fmtos, cls.__mro__)))

    docstrings.keep_types('check_key.parameters', 'kwargs',
                          r'``\*args,\*\*kwargs``')

    @classmethod
    @docstrings.get_sections(base='Plotter._enhance_keys')
    @docstrings.dedent
    def _enhance_keys(cls, keys=None, *args, **kwargs):
        """
        Enhance the given keys by groups

        Parameters
        ----------
        keys: list of str or None
            If None, the all formatoptions of the given class are used. Group
            names from the :attr:`psyplot.plotter.groups` mapping are replaced
            by the formatoptions

        Other Parameters
        ----------------
        %(check_key.parameters.kwargs)s

        Returns
        -------
        list of str
            The enhanced list of the formatoptions"""
        all_keys = list(cls._get_formatoptions())
        if isinstance(keys, six.string_types):
            keys = [keys]
        else:
            keys = list(keys or sorted(all_keys))
        fmto_groups = defaultdict(list)
        for key in all_keys:
            fmto_groups[getattr(cls, key).group].append(key)
        new_i = 0
        for i, key in enumerate(keys[:]):

            if key in fmto_groups:
                del keys[new_i]
                for key2 in fmto_groups[key]:
                    if key2 not in keys:
                        keys.insert(new_i, key2)
                        new_i += 1
            else:
                valid, similar, message = check_key(
                    key, all_keys, False, 'formatoption keyword', *args,
                    **kwargs)
                if not valid:
                    keys.remove(key)
                    new_i -= 1
                    warn(message)
            new_i += 1
        return keys

    @classmethod
    @docstrings.get_sections(base=
        'Plotter.show_keys', sections=['Parameters', 'Returns',
                                       'Other Parameters'])
    @docstrings.dedent
    def show_keys(cls, keys=None, indent=0, grouped=False, func=None,
                  include_links=False, *args, **kwargs):
        """
        Classmethod to return a nice looking table with the given formatoptions

        Parameters
        ----------
        %(Plotter._enhance_keys.parameters)s
        indent: int
            The indentation of the table
        grouped: bool, optional
            If True, the formatoptions are grouped corresponding to the
            :attr:`Formatoption.groupname` attribute

        Other Parameters
        ----------------
        func: function or None
            The function the is used for returning (by default it is printed
            via the :func:`print` function or (when using the gui) in the
            help explorer). The given function must take a string as argument
        include_links: bool or None, optional
            Default False. If True, links (in restructured formats) are
            included in the description. If None, the behaviour is determined
            by the :attr:`psyplot.plotter.Plotter.include_links` attribute.
        %(Plotter._enhance_keys.other_parameters)s

        Returns
        -------
        results of `func`
            None if `func` is the print function, otherwise anything else

        See Also
        --------
        show_summaries, show_docs"""
        def titled_group(groupname):
            bars = str_indent + '*' * len(groupname) + '\n'
            return bars + str_indent + groupname + '\n' + bars

        keys = cls._enhance_keys(keys, *args, **kwargs)
        str_indent = " " * indent
        func = func or default_print_func
        # call this function recursively when grouped is True
        if grouped:
            grouped_keys = DefaultOrderedDict(list)
            for fmto in map(lambda key: getattr(cls, key), keys):
                grouped_keys[fmto.groupname].append(fmto.key)
            text = ""
            for group, keys in six.iteritems(grouped_keys):
                text += titled_group(group) + cls.show_keys(
                    keys, indent=indent, grouped=False, func=six.text_type,
                    include_links=include_links) + '\n\n'
            return func(text.rstrip())

        if not keys:
            return
        n = len(keys)
        ncols = min([4, n])  # number of columns
        # The number of cells in the table is one of the following cases:
        #     1. The number of columns and equal to the number of keys
        #     2. The number of keys
        #     3. The number of keys plus the empty cells in the last column
        ncells = n + ((ncols - (n % ncols)) if n != ncols else 0)
        if include_links or (include_links is None and cls.include_links):
            long_keys = list(map(lambda key: ':attr:`~%s.%s.%s`' % (
                cls.__module__, cls.__name__, key), keys))
        else:
            long_keys = keys
        maxn = max(map(len, long_keys))  # maximal lenght of the keys
        # extend with empty cells
        long_keys.extend([' ' * maxn] * (ncells - n))
        bars = (str_indent + '+-' + ("-"*(maxn) + "-+-")*ncols)[:-1]
        lines = ('| %s |\n%s' % (' | '.join(
            key.ljust(maxn) for key in long_keys[i:i+ncols]), bars)
            for i in range(0, n, ncols))
        text = bars + "\n" + str_indent + ("\n" + str_indent).join(
            lines)
        if six.PY2:
            text = text.encode('utf-8')

        return func(text)

    @classmethod
    @docstrings.dedent
    def _show_doc(cls, fmt_func, keys=None, indent=0, grouped=False,
                  func=None, include_links=False, *args, **kwargs):
        """
        Classmethod to print the formatoptions and their documentation

        This function is the basis for the :meth:`show_summaries` and
        :meth:`show_docs` methods

        Parameters
        ----------
        fmt_func: function
            A function that takes the key, the key as it is printed, and the
            documentation of a formatoption as argument and returns what shall
            be printed
        %(Plotter.show_keys.parameters)s

        Other Parameters
        ----------------
        %(Plotter.show_keys.other_parameters)s

        Returns
        -------
        %(Plotter.show_keys.returns)s

        See Also
        --------
        show_summaries, show_docs"""
        def titled_group(groupname):
            bars = str_indent + '*' * len(groupname) + '\n'
            return bars + str_indent + groupname + '\n' + bars

        func = func or default_print_func

        keys = cls._enhance_keys(keys, *args, **kwargs)
        str_indent = " " * indent
        if grouped:
            grouped_keys = DefaultOrderedDict(list)
            for fmto in map(lambda key: getattr(cls, key), keys):
                grouped_keys[fmto.groupname].append(fmto.key)
            text = "\n\n".join(
                titled_group(group) + cls._show_doc(
                    fmt_func, keys, indent=indent, grouped=False,
                    func=str, include_links=include_links)
                for group, keys in six.iteritems(grouped_keys))
            return func(text.rstrip())

        if include_links or (include_links is None and cls.include_links):
            long_keys = list(map(lambda key: ':attr:`~%s.%s.%s`' % (
                cls.__module__, cls.__name__, key), keys))
        else:
            long_keys = keys

        text = '\n'.join(str_indent + long_key + '\n' + fmt_func(
            key, long_key, getattr(cls, key).__doc__) for long_key, key in zip(
                long_keys, keys))
        return func(text)

    @classmethod
    @docstrings.dedent
    def show_summaries(cls, keys=None, indent=0, *args, **kwargs):
        """
        Classmethod to print the summaries of the formatoptions

        Parameters
        ----------
        %(Plotter.show_keys.parameters)s

        Other Parameters
        ----------------
        %(Plotter.show_keys.other_parameters)s

        Returns
        -------
        %(Plotter.show_keys.returns)s

        See Also
        --------
        show_keys, show_docs"""
        def find_summary(key, key_txt, doc):
            return '\n'.join(wrapper.wrap(doc[:doc.find('\n\n')]))
        str_indent = " " * indent
        wrapper = TextWrapper(width=80, initial_indent=str_indent + ' ' * 4,
                              subsequent_indent=str_indent + ' ' * 4)
        return cls._show_doc(find_summary, keys=keys, indent=indent,
                             *args, **kwargs)

    @classmethod
    @docstrings.dedent
    def show_docs(cls, keys=None, indent=0, *args, **kwargs):
        """
        Classmethod to print the full documentations of the formatoptions

        Parameters
        ----------
        %(Plotter.show_keys.parameters)s

        Other Parameters
        ----------------
        %(Plotter.show_keys.other_parameters)s

        Returns
        -------
        %(Plotter.show_keys.returns)s

        See Also
        --------
        show_keys, show_docs"""
        def full_doc(key, key_txt, doc):
            return ('=' * len(key_txt)) + '\n' + doc + '\n'
        return cls._show_doc(full_doc, keys=keys, indent=indent,
                             *args, **kwargs)

    @classmethod
    def _get_rc_strings(cls):
        """
        Recursive method to get the base strings in the rcParams dictionary.

        This method takes the :attr:`_rcparams_string` attribute from the given
        `class` and combines it with the :attr:`_rcparams_string` attributes
        from the base classes.
        The returned frozenset can be used as base strings for the
        :meth:`psyplot.config.rcsetup.RcParams.find_and_replace` method.

        Returns
        -------
        list
            The first entry is the :attr:`_rcparams_string` of this class,
            the following the :attr:`_rcparams_string` attributes of the
            base classes according to the method resolution order of this
            class"""
        return list(unique_everseen(chain(
            *map(lambda base: getattr(base, '_rcparams_string', []),
                 cls.__mro__))))

    def _set_rc(self):
        """Method to set the rcparams and defaultParams for this plotter"""
        base_str = self._get_rc_strings()
        # to make sure that the '.' is not interpreted as a regex pattern,
        # we specify the pattern_base by ourselves
        pattern_base = map(lambda s: s.replace('.', r'\.'), base_str)
        # pattern for valid keys being all formatoptions in this plotter
        pattern = '(%s)(?=$)' % '|'.join(self._get_formatoptions())
        self._rc = rcParams.find_and_replace(base_str, pattern=pattern,
                                             pattern_base=pattern_base)
        user_rc = SubDict(rcParams['plotter.user'], base_str, pattern=pattern,
                          pattern_base=pattern_base)
        self._rc.update(user_rc.data)

        self._defaultParams = SubDict(rcParams.defaultParams, base_str,
                                      pattern=pattern,
                                      pattern_base=pattern_base)

    docstrings.keep_params('InteractiveBase.update.parameters', 'auto_update')

    @docstrings.dedent
    def update(self, fmt={}, replot=False, auto_update=False, draw=None,
               force=False, todefault=False, **kwargs):
        """
        Update the formatoptions and the plot

        If the :attr:`data` attribute of this plotter is None, the plotter is
        updated like a usual dictionary (see :meth:`dict.update`). Otherwise
        the update is registered and the plot is updated if `auto_update` is
        True or if the :meth:`start_update` method is called (see below).

        Parameters
        ----------
        %(Plotter._register_update.parameters)s
        %(InteractiveBase.start_update.parameters)s
        %(InteractiveBase.update.parameters.auto_update)s
        ``**kwargs``
            Any other formatoption that shall be updated (additionally to those
            in `fmt`)

        Notes
        -----
        %(InteractiveBase.update.notes)s"""
        if self.disabled:
            return
        fmt = dict(fmt)
        if kwargs:
            fmt.update(kwargs)
        # if the data is None, update like a usual dictionary (but with
        # validation)
        if not self._initialized:
            for key, val in six.iteritems(fmt):
                self[key] = val
            return

        self._register_update(fmt=fmt, replot=replot, force=force,
                              todefault=todefault)
        if not self.no_auto_update or auto_update:
            self.start_update(draw=draw)

    def _set_sharing_keys(self, keys):
        """
        Set the keys to share or unshare

        Parameters
        ----------
        keys: string or iterable of strings
            The iterable may contain formatoptions that shall be shared (or
            unshared), or group names of formatoptions to share all
            formatoptions of that group (see the :attr:`fmt_groups` property).
            If None, all formatoptions of this plotter are inserted.

        Returns
        -------
        set
            The set of formatoptions to share (or unshare)"""
        if isinstance(keys, str):
            keys = {keys}
        keys = set(self) if keys is None else set(keys)
        fmto_groups = self._fmto_groups
        keys.update(chain(*(map(lambda fmto: fmto.key, fmto_groups[key])
                            for key in keys.intersection(fmto_groups))))
        keys.difference_update(fmto_groups)
        return keys

    @docstrings.get_sections(base='Plotter.share')
    @docstrings.dedent
    def share(self, plotters, keys=None, draw=None, auto_update=False):
        """
        Share the formatoptions of this plotter with others

        This method shares the formatoptions of this :class:`Plotter` instance
        with others to make sure that, if the formatoption of this changes,
        those of the others change as well

        Parameters
        ----------
        plotters: list of :class:`Plotter` instances or a :class:`Plotter`
            The plotters to share the formatoptions with
        keys: string or iterable of strings
            The formatoptions to share, or group names of formatoptions to
            share all formatoptions of that group (see the
            :attr:`fmt_groups` property). If None, all formatoptions of this
            plotter are unshared.
        %(InteractiveBase.start_update.parameters.draw)s
        %(InteractiveBase.update.parameters.auto_update)s

        See Also
        --------
        unshare, unshare_me"""
        auto_update = auto_update or not self.no_auto_update
        if isinstance(plotters, Plotter):
            plotters = [plotters]
        keys = self._set_sharing_keys(keys)
        for plotter in plotters:
            for key in keys:
                fmto = self._shared.get(key, getattr(self, key))
                if not getattr(plotter, key) == fmto:
                    plotter._shared[key] = getattr(self, key)
                    fmto.shared.add(getattr(plotter, key))
        # now exit if we are not initialized
        if self._initialized:
            self.update(force=keys, auto_update=auto_update, draw=draw)
        for plotter in plotters:
            if not plotter._initialized:
                continue
            old_registered = plotter._registered_updates.copy()
            plotter._registered_updates.clear()
            try:
                plotter.update(force=keys, auto_update=auto_update, draw=draw)
            except:
                raise
            finally:
                plotter._registered_updates.clear()
                plotter._registered_updates.update(old_registered)
        if draw is None:
            draw = rcParams['auto_draw']
        if draw:
            self.draw()
            if rcParams['auto_show']:
                self.show()

    @docstrings.dedent
    def unshare(self, plotters, keys=None, auto_update=False, draw=None):
        """
        Close the sharing connection of this plotter with others

        This method undoes the sharing connections made by the :meth:`share`
        method and releases the given `plotters` again, such that the
        formatoptions in this plotter may be updated again to values different
        from this one.

        Parameters
        ----------
        plotters: list of :class:`Plotter` instances or a :class:`Plotter`
            The plotters to release
        keys: string or iterable of strings
            The formatoptions to unshare, or group names of formatoptions to
            unshare all formatoptions of that group (see the
            :attr:`fmt_groups` property). If None, all formatoptions of this
            plotter are unshared.
        %(InteractiveBase.start_update.parameters.draw)s
        %(InteractiveBase.update.parameters.auto_update)s

        See Also
        --------
        share, unshare_me"""
        auto_update = auto_update or not self.no_auto_update
        if isinstance(plotters, Plotter):
            plotters = [plotters]
        keys = self._set_sharing_keys(keys)
        for plotter in plotters:
            plotter.unshare_me(keys, auto_update=auto_update, draw=draw,
                               update_other=False)
        self.update(force=keys, auto_update=auto_update, draw=draw)

    @docstrings.get_sections(base='Plotter.unshare_me')
    @docstrings.dedent
    def unshare_me(self, keys=None, auto_update=False, draw=None,
                   update_other=True):
        """
        Close the sharing connection of this plotter with others

        This method undoes the sharing connections made by the :meth:`share`
        method and release this plotter again.

        Parameters
        ----------
        keys: string or iterable of strings
            The formatoptions to unshare, or group names of formatoptions to
            unshare all formatoptions of that group (see the
            :attr:`fmt_groups` property). If None, all formatoptions of this
            plotter are unshared.
        %(InteractiveBase.start_update.parameters.draw)s
        %(InteractiveBase.update.parameters.auto_update)s

        See Also
        --------
        share, unshare"""
        auto_update = auto_update or not self.no_auto_update
        keys = self._set_sharing_keys(keys)
        to_update = []
        for key in keys:
            fmto = getattr(self, key)
            try:
                other_fmto = self._shared.pop(key)
            except KeyError:
                pass
            else:
                other_fmto.shared.remove(fmto)
                if update_other:
                    other_fmto.plotter._register_update(
                        force=[other_fmto.key])
                    to_update.append(other_fmto.plotter)
        self.update(force=keys, draw=draw, auto_update=auto_update)
        if update_other and auto_update:
            for plotter in to_update:
                plotter.start_update(draw=draw)

    def get_vfunc(self, key):
        """Return the validation function for a specified formatoption

        Parameters
        ----------
        key: str
            Formatoption key in the :attr:`rc` dictionary

        Returns
        -------
        function
            Validation function for this formatoption"""
        return self._defaultParams[key][1]

    def _save_state(self):
        """Saves the current formatoptions"""
        self._old_fmt.append(self.changed)

    def show(self):
        """Shows all open figures"""
        import matplotlib.pyplot as plt
        plt.show(block=False)

    @dedent
    def has_changed(self, key, include_last=True):
        """
        Determine whether a formatoption changed in the last update

        Parameters
        ----------
        key: str
            A formatoption key contained in this plotter
        include_last: bool
            if True and the formatoption has been included in the last update,
            the return value will not be None. Otherwise the return value will
            only be not None if it changed during the last update

        Returns
        -------
        None or list
            - None, if the value has not been changed during the last update or
              `key` is not a valid formatoption key
            - a list of length two with the old value in the first place and
              the given `value` at the second"""
        if self._initializing or key not in self:
            return
        fmto = getattr(self, key)
        if self._old_fmt and key in self._old_fmt[-1]:
            old_val = self._old_fmt[-1][key]
        else:
            old_val = fmto.default
        if (fmto.diff(old_val) or (include_last and
                                   fmto.key in self._last_update)):
            return [old_val, fmto.value]

    def get_enhanced_attrs(self, arr, axes=['x', 'y', 't', 'z']):
        if isinstance(arr, InteractiveList):
            all_attrs = list(starmap(self.get_enhanced_attrs, zip(
                arr, repeat(axes))))
            attrs = {key: val for key, val in six.iteritems(all_attrs[0])
                     if all(key in attrs and attrs[key] == val
                            for attrs in all_attrs[1:])}
            attrs.update(arr.attrs)
        else:
            attrs = arr.attrs.copy()
            base_variables = self.base_variables
            if len(base_variables) > 1:  # multiple variables
                for name, base_var in six.iteritems(base_variables):
                    attrs.update(
                        {six.text_type(name)+key: value
                         for key, value in six.iteritems(base_var.attrs)})
            else:
                base_var = next(six.itervalues(base_variables))
            attrs['name'] = arr.name
            for dim, coord in six.iteritems(getattr(arr, 'coords', {})):
                if coord.size == 1:
                    attrs[dim] = format_time(coord.values)
            if isinstance(self.data, InteractiveList):
                decoder = self.data[0].psy.decoder
            else:
                decoder = self.data.psy.decoder
            for dim in axes:
                for obj in [base_var, arr]:
                    coord = getattr(decoder, 'get_' + dim)(
                        obj, coords=getattr(arr, 'coords', None))
                    if coord is None:
                        continue
                    if coord.size == 1:
                        attrs[dim] = format_time(coord.values)
                    attrs[dim + 'name'] = coord.name
                    for key, val in six.iteritems(coord.attrs):
                        attrs[dim + key] = val
        self._enhanced_attrs = attrs
        return attrs

    def _make_plot(self):
        plot_fmtos = [fmto for fmto in self._fmtos if fmto.plot_fmt]
        plot_fmtos.sort(key=lambda fmto: fmto.priority, reverse=True)
        for fmto in plot_fmtos:
            self.logger.debug("Making plot with %s formatoption", fmto.key)
            fmto.make_plot()

    @classmethod
    def _get_sample_projection(cls):
        """Returns None. May be subclassed to return a projection that
        can be used when creating a subplot"""
        pass



"""PLUGIN_NAME: PLUGIN_DESC

Describe your plugin here and do whatever you want. It should at least have a
__version__ attribute to specify the version of the package
"""

__version__ = "PLUGIN_VERSION"



# coding: utf-8
"""Warning module of the psyplot python module

This module controls the warning behaviour of the module via the python
builtin warnings module and introduces three new warning classes:

..autosummay::

    PsPylotRuntimeWarning
    PsyPlotWarning
    PsyPlotCritical"""
import warnings
import logging


# disable a warning about "comparison to 'None' in backend_pdf which occurs
# in the matplotlib.backends.backend_pdf.PdfPages class
warnings.filterwarnings(
    'ignore', 'comparison', FutureWarning, 'matplotlib.backends.backend_pdf',
    2264)
# disable a warning about "np.array_split" that occurs for certain numpy
# versions
warnings.filterwarnings(
    'ignore', 'in the future np.array_split will retain', FutureWarning,
    'numpy.lib.shape_base', 431)
# disable a warning about "elementwise comparison of a string" in the
# matplotlib.collection.Collection.get_edgecolor method that occurs for certain
# matplotlib and numpy versions
warnings.filterwarnings(
    'ignore', 'elementwise comparison failed', FutureWarning,
    'matplotlib.collections', 590)


logger = logging.getLogger(__name__)


class PsyPlotRuntimeWarning(RuntimeWarning):
    """Runtime warning that appears only ones"""
    pass


class PsyPlotWarning(UserWarning):
    """Normal UserWarning for psyplot module"""
    pass


class PsyPlotCritical(UserWarning):
    """Critical UserWarning for psyplot module"""
    pass


warnings.simplefilter('always', PsyPlotWarning, append=True)
warnings.simplefilter('always', PsyPlotCritical, append=True)


def disable_warnings(critical=False):
    """Function that disables all warnings and all critical warnings (if
    critical evaluates to True) related to the psyplot Module.
    Please note that you can also configure the warnings via the
    psyplot.warning logger (logging.getLogger(psyplot.warning))."""
    warnings.filterwarnings('ignore', '\w', PsyPlotWarning, 'psyplot', 0)
    if critical:
        warnings.filterwarnings('ignore', '\w', PsyPlotCritical, 'psyplot', 0)


def warn(message, category=PsyPlotWarning, logger=None):
    """wrapper around the warnings.warn function for non-critical warnings.
    logger may be a logging.Logger instance"""
    if logger is not None:
        message = "[Warning by %s]\n%s" % (logger.name, message)
    warnings.warn(message, category, stacklevel=3)


def critical(message, category=PsyPlotCritical, logger=None):
    """wrapper around the warnings.warn function for critical warnings.
    logger may be a logging.Logger instance"""
    if logger is not None:
        message = "[Critical warning by %s]\n%s" % (logger.name, message)
    warnings.warn(message, category, stacklevel=2)


old_showwarning = warnings.showwarning


def customwarn(message, category, filename, lineno, *args, **kwargs):
    """Use the psyplot.warning logger for categories being out of
    PsyPlotWarning and PsyPlotCritical and the default warnings.showwarning
    function for all the others."""
    if category is PsyPlotWarning:
        logger.warning(warnings.formatwarning(
            "\n%s" % message, category, filename, lineno))
    elif category is PsyPlotCritical:
        logger.critical(warnings.formatwarning(
            "\n%s" % message, category, filename, lineno),
            exc_info=True)
    else:
        old_showwarning(message, category, filename, lineno, *args, **kwargs)


warnings.showwarning = customwarn



"""PLUGIN_NAME psyplot plugin

This module defines the rcParams for the PLUGIN_NAME plugin. This module will
be imported when psyplot is imported. What is should contain is:

- an rcParams variable as instance of :class:`psyplot.config.rcsetup.RcParams`
  that describes the configuration of your plugin
- a get_versions function that returns the version of your plugin and the ones
  from its requirements

.. warning::

    Because of recursion issues, You have to load the psyplot module before
    loading this module! In other words, you have to type

    .. code-block:: python

        import psyplot
        import PLUGIN_PYNAME.plugin"""
from psyplot.config.rcsetup import RcParams
from PLUGIN_PYNAME import __version__ as plugin_version


def get_versions(requirements=True):
    """Get the versions of PLUGIN_NAME and it's requirements

    Parameters
    ----------
    requirements: bool
        If True, the requirements are imported and it's versions are included
    """
    ret = {'version': plugin_version}
    if requirements:
        # insert versions of the requirements, e.g. via
        #   >>> import requirement
        #   >>> ret['requirement'] = requirement.__version__
        pass
    return ret


# -----------------------------------------------------------------------------
# ------------------------- validation functions ------------------------------
# -----------------------------------------------------------------------------


# define your validation functions for the values in the rcParams here. If
# a validation fails, the function should raise a ValueError or TypeError


# -----------------------------------------------------------------------------
# ------------------------------ rcParams -------------------------------------
# -----------------------------------------------------------------------------


# define your defaultParams. A mapping from rcParams key to a list of length 3:
#
# 1. the default value
# 2. the validation function of type conversion function
# 3. a short description of the default value
#
# Example::
#
#     defaultParams = {'my.key': [True, bool, 'What my key does']}
defaultParams = {

    # key for defining new plotters
    'project.plotters': [
        {'plot_method_identifer': {
             'module': 'PLUGIN_PYNAME.plotters',
             'plotter_name': 'MyPlotter',   # or any other name
             # any other item for the :func:`psyplot.project.register_plotter`
             # function
             # 'plot_func': False,
             # 'prefer_list': True,
             # ...
             },
         }, dict, "The plot methods in the PLUGIN_NAME package"],
    # if you define new plotters, we recommend to assign a specific rcParams
    # key for it, e.g.
    # 'plotter.PLUGIN_PYNAME.my_fmt': [1, int, ' the value for my_fmt']
    }

# create the rcParams and populate them with the defaultParams. For more
# information on this class, see the :class:`psyplot.config.rcsetup.RcParams`
# class
rcParams = RcParams(defaultParams=defaultParams)
rcParams.update_from_defaultParams()



# -*- coding: utf-8 -*-
"""Gdal Store for reading GeoTIFF files into an :class:`xarray.Dataset`

This module contains the definition of the :class:`GdalStore` class that can
be used to read in a GeoTIFF file into an :class:`xarray.Dataset`.
It requires that you have the python gdal module installed.

Examples
--------
to open a GeoTIFF file named ``'my_tiff.tiff'`` you can do::

    >>> from psyplot.gdal_store import GdalStore
    >>> from xarray import open_dataset
    >>> ds = open_dataset(GdalStore('my_tiff'))

Or you use the `engine` of the :func:`psyplot.open_dataset` function:

    >>> ds = open_dataset('my_tiff.tiff', engine='gdal')"""
import six
from numpy import arange, nan, dtype
from xarray import Variable
from collections import OrderedDict
try:
    from xarray.core.utils import FrozenOrderedDict
except ImportError:
    FrozenOrderedDict = dict
from xarray.backends.common import AbstractDataStore
from psyplot.compat.pycompat import range
from psyplot.warning import warn
import psyplot.data as psyd
try:
    import gdal
    from osgeo import gdal_array
except ImportError as e:
    gdal = psyd._MissingModule(e)
try:
    from dask.array import Array
    with_dask = True
except ImportError:
    with_dask = False


class GdalStore(AbstractDataStore):
    """Datastore to read raster files suitable for the gdal package

    We recommend to use the :func:`psyplot.open_dataset` function to open
    a geotiff file::

        >>> ds = psyplot.open_dataset('my_geotiff.tiff', engine='gdal')

    Notes
    -----
    The :class:`GdalStore` object is not as elaborate as, for example, the
    `gdal_translate` command. Many attributes, e.g. variable names or netCDF
    dimensions will not be interpreted. We only support two
    dimensional arrays and each band is saved into one variable named like
    ``'Band1', 'Band2', ...``. If you want a more elaborate translation of your
    GDAL Raster, convert the file to a netCDF file using ``gdal_translate`` or
    the ``gdal.GetDriverByName('netCDF').CreateCopy`` method. However this
    class does not create an extra file on your hard disk as it is done by
    GDAL."""

    def __init__(self, filename_or_obj):
        """
        Parameters
        ----------
        filename_or_obj: str
            The path to the GeoTIFF file or a gdal dataset"""
        if isinstance(psyd.safe_list(filename_or_obj)[0], six.string_types):
            self.ds = gdal.Open(filename_or_obj)
            self._filename = filename_or_obj
        else:
            self.ds = filename_or_obj
            fnames = self.ds.GetFileList()
            self._filename = fnames[0] if len(fnames) == 1 else fnames

    def get_variables(self):
        def load(band):
            band = ds.GetRasterBand(band)
            a = band.ReadAsArray()
            no_data = band.GetNoDataValue()
            if no_data is not None:
                try:
                    a[a == no_data] = a.dtype.type(nan)
                except ValueError:
                    pass
            return a
        ds = self.ds
        dims = ['lat', 'lon']
        chunks = ((ds.RasterYSize,), (ds.RasterXSize,))
        shape = (ds.RasterYSize, ds.RasterXSize)
        variables = OrderedDict()
        for iband in range(1, ds.RasterCount+1):
            band = ds.GetRasterBand(iband)
            dt = dtype(gdal_array.codes[band.DataType])
            if with_dask:
                dsk = {('x', 0, 0): (load, iband)}
                arr = Array(dsk, 'x', chunks, shape=shape, dtype=dt)
            else:
                arr = load(iband)
            attrs = band.GetMetadata_Dict()
            try:
                dt.type(nan)
                attrs['_FillValue'] = nan
            except ValueError:
                no_data = band.GetNoDataValue()
                attrs.update({'_FillValue': no_data} if no_data else {})
            variables['Band%i' % iband] = Variable(dims, arr, attrs)
        variables['lat'], variables['lon'] = self._load_GeoTransform()
        return FrozenOrderedDict(variables)

    def _load_GeoTransform(self):
        """Calculate latitude and longitude variable calculated from the
        gdal.Open.GetGeoTransform method"""
        def load_lon():
            return arange(ds.RasterXSize)*b[1]+b[0]

        def load_lat():
            return arange(ds.RasterYSize)*b[5]+b[3]
        ds = self.ds
        b = self.ds.GetGeoTransform()  # bbox, interval
        if with_dask:
            lat = Array(
                {('lat', 0): (load_lat,)}, 'lat', (self.ds.RasterYSize,),
                shape=(self.ds.RasterYSize,), dtype=float)
            lon = Array(
                {('lon', 0): (load_lon,)}, 'lon', (self.ds.RasterXSize,),
                shape=(self.ds.RasterXSize,), dtype=float)
        else:
            lat = load_lat()
            lon = load_lon()
        return Variable(('lat',), lat), Variable(('lon',), lon)

    def get_attrs(self):
        from osr import SpatialReference
        attrs = self.ds.GetMetadata()
        try:
            sp = SpatialReference(wkt=self.ds.GetProjection())
            proj4 = sp.ExportToProj4()
        except:
            warn('Could not identify projection')
        else:
            attrs['proj4'] = proj4
        return FrozenOrderedDict(attrs)



# -*- coding: utf-8 -*-
import os
import os.path as osp
import sys
import argparse
import pickle
import six
import glob
from itertools import chain
from collections import defaultdict
import yaml
import xarray as xr
import psyplot
from psyplot.docstring import docstrings
from psyplot.warning import warn
from psyplot.compat.pycompat import get_default_value
from funcargparse import FuncArgParser
import logging

rcParams = psyplot.rcParams


logger = logging.getLogger(__name__)


def main(args=None):
    """Main function for usage of psyplot from the command line

    This function creates a parser that parses command lines to the
    :func:`make_plot` functions or (if the ``psyplot_gui`` module is
    present, to the :func:`psyplot_gui.start_app` function)

    Returns
    -------
    psyplot.parser.FuncArgParser
        The parser that has been used from the command line"""
    try:
        from psyplot_gui import get_parser as _get_parser
    except ImportError:
        logger.debug('Failed to import gui', exc_info=True)
        parser = get_parser(create=False)
        parser.update_arg('output', required=True)
        parser.create_arguments()
        parser.parse2func(args)
    else:
        parser = _get_parser(create=False)
        parser.create_arguments()
        parser.parse_known2func(args)


@docstrings.get_sections(base='make_plot')
@docstrings.dedent
def make_plot(fnames=[], name=[], dims=None, plot_method=None,
              output=None, project=None, engine=None, formatoptions=None,
              tight=False, rc_file=None, encoding=None, enable_post=False,
              seaborn_style=None, output_project=None,
              concat_dim=get_default_value(xr.open_mfdataset, 'concat_dim'),
              chname={}, preset=None):
    """
    Eventually start the QApplication or only make a plot

    Parameters
    ----------
    fnames: list of str
        Either the filenames to show, or, if the `project` parameter is set,
        the a list of `,`-separated filenames to make a mapping from the
        original filename to a new one
    name: list of str
        The variable names to plot if the `output` parameter is set
    dims: dict
        A mapping from coordinate names to integers if the `project` is not
        given
    plot_method: str
        The name of the plot_method to use
    output: str or list of str
        If set, the data is loaded and the figures are saved to the specified
        filename and now graphical user interface is shown
    project: str
        If set, the project located at the given file name is loaded
    engine: str
        The engine to use for opening the dataset (see
        :func:`psyplot.data.open_dataset`)
    formatoptions: dict
        A dictionary of formatoption that is applied to the data visualized by
        the chosen `plot_method`
    tight: bool
        If True/set, it is tried to figure out the tight bbox of the figure and
        adjust the paper size of the `output` to it
    rc_file: str
        The path to a yaml configuration file that can be used to update  the
        :attr:`~psyplot.config.rcsetup.rcParams`
    encoding: str
        The encoding to use for loading the project. If None, it is
        automatically determined by pickle. Note: Set this to ``'latin1'``
        if using a project created with python2 on python3.
    enable_post: bool
        Enable the :attr:`~psyplot.plotter.Plotter.post` processing
        formatoption. If True/set, post processing scripts are enabled in the
        given `project`. Only set this if you are sure that you can trust the
        given project file because it may be a security vulnerability.
    seaborn_style: str
        The name of the style of the seaborn package that can be used for
        the :func:`seaborn.set_style` function
    output_project: str
        The name of a project file to save the project to
    concat_dim: str
        The concatenation dimension if multiple files in `fnames` are
        provided
    chname: dict
        A mapping from variable names in the project to variable names in the
        datasets that should be used instead
    preset: str
        The filename or identifier of a preset. If the given `preset` is
        the path to an existing yaml file, it will be loaded. Otherwise we
        look up the `preset` in the psyplot configuration directory (see
        :func:`~psyplot.config.rcsetup.get_configdir`).
    """
    if project is not None and (name != [] or dims is not None):
        warn('The `name` and `dims` parameter are ignored if the `project`'
             ' parameter is set!')
    if rc_file is not None:
        rcParams.load_from_file(rc_file)

    if dims is not None and not isinstance(dims, dict):
        dims = dict(chain(*map(six.iteritems, dims)))

    if len(output) == 1:
        output = output[0]
    if not fnames and not project:
        raise ValueError(
            "Either a filename or a project file must be provided if "
            "the output parameter is set!")
    elif project is None and plot_method is None:
        raise ValueError(
            "A plotting method must be provided if the output parameter "
            "is set and not the project!")
    if seaborn_style is not None:
        import seaborn as sns
        sns.set_style(seaborn_style)
    import psyplot.project as psy
    if project is not None:
        fnames = [s.split(',') for s in fnames]
        chname = dict(chname)
        single_files = (l[0] for l in fnames if len(l) == 1)
        alternative_paths = defaultdict(lambda: next(single_files, None))
        alternative_paths.update([l for l in fnames if len(l) == 2])
        p = psy.Project.load_project(
            project, alternative_paths=alternative_paths,
            engine=engine, encoding=encoding, enable_post=enable_post,
            chname=chname)
        if preset:
            p.load_preset(preset)
        if formatoptions is not None:
            p.update(fmt=formatoptions)
        p.export(output, tight=tight)
    else:
        pm = getattr(psy.plot, plot_method, None)
        if pm is None:
            raise ValueError("Unknown plot method %s!" % plot_method)
        kwargs = {'name': name} if name else {}
        p = pm(
            fnames, dims=dims or {}, engine=engine, preset=preset,
            fmt=formatoptions or {}, mf_mode=True, concat_dim=concat_dim,
            **kwargs)
        p.export(output, tight=tight)
    if output_project is not None:
        p.save_project(output_project)
    return


def get_parser(create=True):
    """Return a parser to make that can be used to make plots or open files
    from the command line

    Returns
    -------
    psyplot.parser.FuncArgParser
        The :class:`argparse.ArgumentParser` instance"""
    #: The parse that is used to parse arguments from the command line
    epilog = docstrings.get_sections(docstrings.dedent("""
        Examples
        --------

        Here are some examples on how to use psyplot from the command line.

        Plot the variable ``'t2m'`` in a netCDF file ``'myfile.nc'`` and save
        the plot to ``'plot.pdf'``::

            $ psyplot myfile.nc -n t2m -pm mapplot -o test.pdf

        Create two plots for ``'t2m'`` with the first and second timestep on
        the second vertical level::

            $ psyplot myfile.nc -n t2m  -pm mapplot -o test.pdf -d t,0,1 z,1

        If you have save a project using the
        :meth:`psyplot.project.Project.save_project` method into a file named
        ``'project.pkl'``, you can replot this via::

            $ psyplot -p project.pkl -o test.pdf

        If you use a different dataset than the one you used in the project
        (e.g. ``'other_ds.nc'``), you can replace it via::

            $ psyplot other_dataset.nc -p project.pkl -o test.pdf

        or explicitly via::

            $ psyplot old_ds.nc,other_ds.nc -p project.pkl -o test.pdf

        You can also load formatoptions from a configuration file, e.g.::

            $ echo 'title: my title' > fmt.yaml
            $ psyplot myfile.nc -n t2m  -pm mapplot -fmt fmt.yaml -o test.pdf
        """), 'parser', ['Examples'])

    if _on_rtd:  # make a rubric examples section
        epilog = '.. rubric:: Examples\n' + '\n'.join(epilog.splitlines()[2:])

    parser = FuncArgParser(
        description="""
        Load a dataset, make the plot and save the result to a file""",
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    info_grp = parser.add_argument_group(
        'Info options',
        'Options that print informations and quit afterwards')

    parser.update_arg('version', short='V', long='version', action='version',
                      version=psyplot.__version__, if_existent=False,
                      group=info_grp)

    parser.update_arg('all_versions', short='aV', long='all-versions',
                      action=AllVersionsAction, if_existent=False,
                      group=info_grp)

    parser.update_arg('list_plugins', short='lp', long='list-plugins',
                      action=ListPluginsAction, if_existent=False,
                      group=info_grp)
    parser.update_arg(
        'list_plot_methods', short='lpm', long='list-plot-methods',
        action=ListPlotMethodsAction, if_existent=False, group=info_grp)
    parser.update_arg(
        'list_datasets', short='lds', long='list-datasets',
        action=ListDsNamesAction, if_existent=False, group=info_grp,
        help="""List the used dataset names in the given `project`.""")

    parser.update_arg(
        'list_presets', short='lps', long='list-presets',
        action=ListPresetsAction, if_existent=False, group=info_grp)

    parser.setup_args(make_plot)

    output_grp = parser.add_argument_group(
        'Output options',
        'Options that only have an effect if the `-o` option is set.')

    parser.update_arg('fnames', positional=True, nargs='*')

    parser.update_arg('name', short='n', nargs='*', metavar='variable_name',
                      const=None)

    parser.update_arg('dims', short='d', nargs='+', type=_load_dims,
                      metavar='dim,val1[,val2[,...]]')

    pm_choices = {pm for pm, d in filter(
                      lambda t: t[1].get('plot_func', True),
                      six.iteritems(rcParams['project.plotters']))}
    if psyplot._project_imported:
        import psyplot.project as psy
        pm_choices.update(set(psy.plot._plot_methods))
    parser.update_arg('plot_method', short='pm', choices=pm_choices,
                      metavar='{%s}' % ', '.join(map(repr, pm_choices)))

    parser.update_arg('output', short='o', group=output_grp)
    parser.update_arg('output_project', short='op', group=output_grp)

    parser.update_arg('project', short='p')

    parser.update_arg(
        'formatoptions', short='fmt', type=_load_dict, help="""
        The path to a yaml (``'.yml'`` or ``'.yaml'``) or pickle file
        defining a dictionary of formatoption that is applied to the data
        visualized by the chosen `plot_method`""", metavar='FILENAME')

    parser.update_arg(
        'chname', type=lambda s: s.split(','), nargs='*', help="""
        A mapping from variable names in the project to variable names in the
        datasets that should be used instead. Variable names should be
        separated by a comma.""", metavar='project-variable,variable-to-use')

    parser.update_arg('tight', short='t', group=output_grp)

    parser.update_arg('rc_file', short='rc')
    parser.pop_key('rc_file', 'metavar')

    parser.update_arg('encoding', short='e')

    parser.pop_key('enable_post', 'short')

    parser.update_arg('seaborn_style', short='sns')

    parser.update_arg('concat_dim', short='cd')

    if create:
        parser.create_arguments()

    return parser


def _load_dict(fname):
    with open(fname) as f:
        if fname.endswith('.yml') or fname.endswith('.yaml'):
            return yaml.load(f, Loader=yaml.SafeLoader)
        return pickle.load(f)


def _load_dims(s):
    s = s.split(',')
    if len(s) > 1:
        return {s[0]: list(map(int, s[1:]))}
    return {}


#: Disable the default for the info actions on RTD, because it looks
#: better in the docs
_on_rtd = os.environ.get('READTHEDOCS', None) == 'True'


class AllVersionsAction(argparse.Action):

    def __init__(self, option_strings, dest=argparse.SUPPRESS, nargs=None,
                 default=argparse.SUPPRESS, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        kwargs['help'] = ("Print the versions of all plugins and requirements "
                          "and exit")
        if not _on_rtd:
            kwargs['default'] = default
        super(AllVersionsAction, self).__init__(
            option_strings, nargs=0, dest=dest,
            **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        print(yaml.dump(psyplot.get_versions(), default_flow_style=False))
        sys.exit(0)


class ListPresetsAction(argparse.Action):

    def __init__(self, option_strings, dest=argparse.SUPPRESS, nargs=None,
                 default=argparse.SUPPRESS, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        kwargs['help'] = ("Print available presets and exit")
        if not _on_rtd:
            kwargs['default'] = default
        super().__init__(option_strings, nargs=0, dest=dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        from psyplot.config.rcsetup import get_configdir
        presets_dir = osp.join(get_configdir(), 'presets')
        if not osp.exists(presets_dir):
            sys.exit(0)
        else:
            presets = {osp.splitext(osp.basename(fname))[0]: fname
                       for fname in glob.glob(osp.join(presets_dir, '*.yml'))}
            print('\n'.join(map(': '.join, presets.items())))
            sys.exit(0)

class ListPluginsAction(argparse.Action):

    def __init__(self, option_strings, dest=argparse.SUPPRESS, nargs=None,
                 default=argparse.SUPPRESS, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        kwargs['help'] = ("Print the names of the plugins and exit")
        if not _on_rtd:
            kwargs['default'] = default
        super(ListPluginsAction, self).__init__(
            option_strings, nargs=0, dest=dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        print(yaml.dump(psyplot.rcParams._plugins, default_flow_style=False))
        sys.exit(0)


class ListPlotMethodsAction(argparse.Action):

    def __init__(self, option_strings, dest=argparse.SUPPRESS, nargs=None,
                 default=argparse.SUPPRESS, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        kwargs['help'] = "List the available plot methods and what they do"
        if not _on_rtd:
            kwargs['default'] = default
        super(ListPlotMethodsAction, self).__init__(
            option_strings, nargs=0, dest=dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        pm_choices = {}
        for pm, d in filter(lambda t: t[1].get('plot_func', True),
                            six.iteritems(rcParams['project.plotters'])):
            pm_choices[pm] = d.get('summary') or (
                'Open and plot data via :class:`%s.%s` plotters' % (
                    d['module'], d['plotter_name']))
        if psyplot._project_imported:
            import psyplot.project as psy
            pm_choices.update(psy.plot._plot_methods)
        print(yaml.dump(pm_choices, default_flow_style=False))
        sys.exit(0)


class ListDsNamesAction(argparse.Action):
    """An action to list the used file names in a project"""

    def __init__(self, option_strings, dest=argparse.SUPPRESS, nargs=None,
                 default=argparse.SUPPRESS, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        if not _on_rtd:
            kwargs['default'] = default
        super(ListDsNamesAction, self).__init__(
            option_strings, nargs=0, dest=dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.project is None:
            print('A project is required before this argument! Call syntax:\n'
                  '%s -p <project-file>.pkl %s' % (parser.prog, option_string))
            sys.exit(1)
        import psyplot.data as psyd
        import pickle
        with open(namespace.project, 'rb') as f:
            d = pickle.load(f)['arrays']
        names = list(filter(None, (
            t[0] for t in psyd.ArrayList._get_dsnames(d))))
        if names:
            print(yaml.dump(names, default_flow_style=False))
        sys.exit(0)


if __name__ == '__main__':
    main()



"""Docstring module of the psyplot package

We use the docrep_ package for managing our docstrings

.. _docrep: http://docrep.readthedocs.io/en/latest/
"""
import types
import six
import inspect
from docrep import DocstringProcessor, safe_modulo


def dedent(func):
    """
    Dedent the docstring of a function and substitute with :attr:`params`

    Parameters
    ----------
    func: function
        function with the documentation to dedent"""
    if isinstance(func, types.MethodType) and not six.PY3:
        func = func.im_func
    func.__doc__ = func.__doc__ and inspect.cleandoc(func.__doc__)
    return func


def indent(text, num=4):
    """Indet the given string"""
    str_indent = ' ' * num
    return str_indent + ('\n' + str_indent).join(text.splitlines())


def append_original_doc(parent, num=0):
    """Return an iterator that append the docstring of the given `parent`
    function to the applied function"""
    def func(func):
        func.__doc__ = func.__doc__ and func.__doc__ + indent(
            parent.__doc__, num)
        return func
    return func


_docstrings = DocstringProcessor()

_docstrings.get_sections(base='DocstringProcessor.get_sections')(
        dedent(DocstringProcessor.get_sections))


class PsyplotDocstringProcessor(DocstringProcessor):
    """
    A :class:`docrep.DocstringProcessor` subclass with possible types section
    """

    param_like_sections = DocstringProcessor.param_like_sections + [
        'Possible types']

    @_docstrings.dedent
    def get_sections(self, s=None, base=None, sections=[
            'Parameters', 'Other Parameters', 'Possible types']):
        """
        Extract the specified sections out of the given string

        The same as the :meth:`docrep.DocstringProcessor.get_sections` method
        but uses the ``'Possible types'`` section by default, too

        Parameters
        ----------
        %(DocstringProcessor.get_sections.parameters)s

        Returns
        -------
        str
            The replaced string
        """
        return super(PsyplotDocstringProcessor, self).get_sections(s, base,
                                                                   sections)

del _docstrings

#: :class:`docrep.PsyplotDocstringProcessor` instance that simplifies the reuse
#: of docstrings from between different python objects.
docstrings = PsyplotDocstringProcessor()



"""Module for creating a new template for a psyplot plugin
"""
import funcargparse
import os
import os.path as osp
import shutil


def new_plugin(odir, py_name=None, version='0.0.1.dev0',
               description='New plugin'):
    """
    Create a new plugin for the psyplot package

    Parameters
    ----------
    odir: str
        The name of the directory where the data will be copied to. The
        directory must not exist! The name of the directory also defines the
        name of the package.
    py_name: str
        The name of the python package. If None, the basename of `odir` is used
        (and ``'-'`` is replaced by ``'_'``)
    version: str
        The version of the package
    description: str
        The description of the plugin"""
    name = osp.basename(odir)
    if py_name is None:
        py_name = name.replace('-', '_')

    src = osp.join(osp.dirname(__file__), 'plugin-template-files')
    # copy the source files
    shutil.copytree(src, odir)
    os.rename(osp.join(odir, 'plugin_template'), osp.join(odir, py_name))

    replacements = {'PLUGIN_NAME': name,
                    'PLUGIN_PYNAME': py_name,
                    'PLUGIN_VERSION': version,
                    'PLUGIN_DESC': description,
                    }
    files = [
        'README.md',
        'setup.py',
        osp.join(py_name, 'plugin.py'),
        osp.join(py_name, 'plotters.py'),
        osp.join(py_name, '__init__.py'),
        ]

    for fname in files:
        with open(osp.join(odir, fname)) as f:
            s = f.read()
        for key, val in replacements.items():
            s = s.replace(key, val)
        with open(osp.join(odir, fname), 'w') as f:
            f.write(s)


def main(args=None):
    parser = funcargparse.FuncArgParser()
    parser.setup_args(new_plugin)
    parser.update_short(version='v', description='d')
    parser.create_arguments()
    parser.parse2func(args)


if __name__ == '__main__':
    main()



"""Configuration module of the psyplot package

This module contains the module for managing rc parameters and the logging.
Default parameters are defined in the :data:`rcsetup.defaultParams`
dictionary, however you can set up your own configuration in a yaml file (see
:func:`psyplot.load_rc_from_file`)"""
from .logsetup import setup_logging

#: :class:`str`. Path to the yaml logging configuration file
logcfg_path = setup_logging()


from .rcsetup import psyplot_fname


#: class:`str` or ``None``. Path to the yaml configuration file (if found).
#: See :func:`~psyplot.config.rcsetup.psyplot_fname` for further information
config_path = psyplot_fname()



"""Sphinx extension module to provide additional sections for numpy docstrings

This extension extends the :mod:`sphinx.ext.napoleon` package with an
additional *Possible types* section in order to document possible types for
descriptors.

Notes
-----
If you use this module as a sphinx extension, you should not list the
:mod:`sphinx.ext.napoleon` module in the extensions variable of your conf.py.
This module has been tested for sphinx 1.3.1."""
from abc import ABCMeta, abstractmethod
from sphinx.ext.napoleon import (
    NumpyDocstring, GoogleDocstring, setup as napoleon_setup)


class DocstringExtension(object):
    """Class that introduces a "Possible Types" section

    This class serves as a base class for
    :class:`sphinx.ext.napoleon.NumpyDocstring` and
    :class:`sphinx.ext.napoleon.GoogleDocstring` to introduce
    another section names *Possible types*

    Examples
    --------
    The usage is the same as for the NumpyDocstring class, but it supports
    the `Possible types` section::

        >>> from sphinx.ext.napoleon import Config

        >>> from psyplot.sphinxext.extended_napoleon import (
        ...     ExtendedNumpyDocstring)
        >>> config = Config(napoleon_use_param=True,
        ...                 napoleon_use_rtype=True)
        >>> docstring = '''
        ... Possible types
        ... --------------
        ... type1
        ...     Description of `type1`
        ... type2
        ...     Description of `type2`'''
        >>> print(ExtendedNumpyDocstring(docstring, config))
        .. rubric:: Possible types

        * *type1* --
          Description of `type1`
        * *type2* --
          Description of `type2`"""
    __metaclass__ = ABCMeta

    def _parse_possible_types_section(self, section):
        fields = self._consume_fields(prefer_type=True)
        lines = ['.. rubric:: %s' % section, '']
        multi = len(fields) > 1
        for _name, _type, _desc in fields:
            field = self._format_field(_name, _type, _desc)
            if multi:
                lines.extend(self._format_block('* ', field))
            else:
                lines.extend(field)
        return lines + ['']

    @abstractmethod
    def _parse(self):
        pass


class ExtendedNumpyDocstring(NumpyDocstring, DocstringExtension):
    """:class:`sphinx.ext.napoleon.NumpyDocstring` with more sections"""

    def _parse(self, *args, **kwargs):
        self._sections['possible types'] = self._parse_possible_types_section
        return super(ExtendedNumpyDocstring, self)._parse(*args, **kwargs)


class ExtendedGoogleDocstring(GoogleDocstring, DocstringExtension):
    """:class:`sphinx.ext.napoleon.GoogleDocstring` with more sections"""
    def _parse(self, *args, **kwargs):
        self._sections['possible types'] = self._parse_possible_types_section
        return super(ExtendedGoogleDocstring, self)._parse(*args, **kwargs)


def process_docstring(app, what, name, obj, options, lines):
    """Process the docstring for a given python object.

    Called when autodoc has read and processed a docstring. `lines` is a list
    of docstring lines that `_process_docstring` modifies in place to change
    what Sphinx outputs.

    The following settings in conf.py control what styles of docstrings will
    be parsed:

    * ``napoleon_google_docstring`` -- parse Google style docstrings
    * ``napoleon_numpy_docstring`` -- parse NumPy style docstrings

    Parameters
    ----------
    app : sphinx.application.Sphinx
        Application object representing the Sphinx process.
    what : str
        A string specifying the type of the object to which the docstring
        belongs. Valid values: "module", "class", "exception", "function",
        "method", "attribute".
    name : str
        The fully qualified name of the object.
    obj : module, class, exception, function, method, or attribute
        The object to which the docstring belongs.
    options : sphinx.ext.autodoc.Options
        The options given to the directive: an object with attributes
        inherited_members, undoc_members, show_inheritance and noindex that
        are True if the flag option of same name was given to the auto
        directive.
    lines : list of str
        The lines of the docstring, see above.

        .. note:: `lines` is modified *in place*

    Notes
    -----
    This function is (to most parts) taken from the :mod:`sphinx.ext.napoleon`
    module, sphinx version 1.3.1, and adapted to the classes defined here"""
    result_lines = lines
    if app.config.napoleon_numpy_docstring:
        docstring = ExtendedNumpyDocstring(
            result_lines, app.config, app, what, name, obj, options)
        result_lines = docstring.lines()
    if app.config.napoleon_google_docstring:
        docstring = ExtendedGoogleDocstring(
            result_lines, app.config, app, what, name, obj, options)
        result_lines = docstring.lines()

    lines[:] = result_lines[:]


def setup(app):
    """Sphinx extension setup function

    When the extension is loaded, Sphinx imports this module and executes
    the ``setup()`` function, which in turn notifies Sphinx of everything
    the extension offers.

    Parameters
    ----------
    app : sphinx.application.Sphinx
        Application object representing the Sphinx process

    Notes
    -----
    This function uses the setup function of the :mod:`sphinx.ext.napoleon`
    module"""
    from sphinx.application import Sphinx
    if not isinstance(app, Sphinx):
        return  # probably called by tests

    app.connect('autodoc-process-docstring', process_docstring)
    return napoleon_setup(app)



"""Setup file for plugin PLUGIN_NAME

This file is used to install the package to your python distribution.
Installation goes simply via::

    python setup.py install
"""

from setuptools import setup, find_packages


setup(name='PLUGIN_NAME',
      version='PLUGIN_VERSION',
      description='PLUGIN_DESC',
      keywords='visualization psyplot',
      license="GPLv2",
      packages=find_packages(exclude=['docs', 'tests*', 'examples']),
      install_requires=[
          'psyplot',
      ],
      entry_points={'psyplot': ['plugin=PLUGIN_PYNAME.plugin']},
      zip_safe=False)



"""Compatibility module for different python versions"""
import os
import six
import inspect

if six.PY3:

    class DictMethods(object):

        @staticmethod
        def iteritems(d):
            return iter(dict.items(d))

        @staticmethod
        def itervalues(d):
            return iter(dict.values(d))

        @staticmethod
        def iterkeys(d):
            return iter(dict.keys(d))

    def getcwd(*args, **kwargs):
        return os.getcwd(*args, **kwargs)

    def get_default_value(func, arg):
        argspec = inspect.getfullargspec(func)
        return next(default for a, default in zip(reversed(argspec[0]),
                                                  reversed(argspec.defaults))
                    if a == arg)

    basestring = str
    unicode_type = str
    bytes_type = bytes
    range = range
    zip = zip
    filter = filter
    map = map
    from functools import reduce
    from itertools import filterfalse
    import builtins
    from queue import Queue

elif six.PY2:
    # Python 2

    class DictMethods(object):
        """okay"""
        @staticmethod
        def iteritems(d):
            "checked"
            return dict.iteritems(d)

        @staticmethod
        def itervalues(d):
            return dict.itervalues(d)

        @staticmethod
        def iterkeys(d):
            return dict.iterkeys(d)

    def getcwd(*args, **kwargs):
        return os.getcwdu(*args, **kwargs)

    def get_default_value(func, arg):
        argspec = inspect.getargspec(func)
        return next(default for a, default in zip(reversed(argspec[0]),
                                                  reversed(argspec.defaults))
                    if a == arg)

    basestring = basestring
    unicode_type = unicode
    bytes_type = str
    range = xrange
    from itertools import (izip as zip, imap as map, ifilter as filter,
                           ifilterfalse as filterfalse)
    reduce = reduce
    import __builtin__ as builtins
    from Queue import Queue

try:
    from cyordereddict import OrderedDict
except ImportError:
    try:
        from collections import OrderedDict
    except ImportError:
        from ordereddict import OrderedDict

try:
    from collections import UserDict
except ImportError:
    from UserDict import IterableUserDict as UserDict


def isstring(s):
    return isinstance(s, six.string_types)
